# coding: utf-8
"""
SCRIPT NAME:    misc_deterministic.py
AUTHOR:         YAN FU
DATE:           30 July 2018
DESCRIPTION:    This code performs time equivalence analysis without probabilistic component.
"""

# DEPENDENCIES
import sys
import copy
import os
import json
import numpy as np
from sfeprapy import time_equivalence_mc as func_core
from sfeprapy.func.temperature_fires import standard_fire_iso834 as _fire_standard
from sfeprapy.dat.steel_carbon import Thermal


def main(dir_wd_list=list()):

    # PARSE INPUT FILES DIRECTORY
    # ===========================

    # if no function parameters
    if len(sys.argv) > 1 > len(dir_wd_list):
        dir_wd_list = copy.copy(sys.argv)
        del dir_wd_list[0]

    else:
        while True:
            project_full_path = input("Work directory (blank to finish): ")
            if len(project_full_path) == 0:
                break
            else:
                dir_wd_list.append(project_full_path)

    # exit function if no work directory obtained
    if len(dir_wd_list) < 1:
        raise ValueError("Aborted. No directories obtained.")

    # otherwise, transform path to absolute real path
    else:
        dir_wd_list = [os.path.abspath(os.path.realpath(i)) for i in dir_wd_list]

    # PARSE INPUT FILE PATHS
    # ======================
    files_directories = []
    for d in dir_wd_list:

        list_files = os.listdir(d)
        list_input_files = []
        for f in list_files:
            if f.endswith('.json'):
                list_input_files.append(os.path.join(d, f))

        files_directories.append(list_input_files)

    # TIME EQUIVALENCE CALCULATION
    # ============================

    # "iso834_time": 0,
    # "iso834_temperature": 0,
    steel_prop = Thermal()
    beam_c = steel_prop.c()
    fire = _fire_standard(np.arange(0, 3*60*60, 1), 273.15+20)
    # iterate through all input work directories
    print(files_directories)
    dirs_calcs = []
    for d in files_directories:

        # iterate all input files in the work directory
        calcs = []
        for f in d:
            with open(str(f), "r") as i:
                kwargs = json.load(i)
                kwargs["iso834_time"] = fire[0]
                kwargs["iso834_temperature"] = fire[1]
                kwargs["beam_c"] = beam_c
                kwargs["is_return_dict"] = True

            c = func_core.calc_time_equivalence(**kwargs)

            calcs.append(func_core.calc_time_equivalence(**kwargs))

            with open(f + ".out.json", 'w') as f_:
                json.dump(c, f_, indent=True)

            print('-'*40)
            print("{:<30}{:<10}".format("Key", 'Value'))
            print('-'*40)
            for k, v in c.items():
                print("{:<30}{:<10}".format(k, v))
            print('-'*40)

        dirs_calcs.append(calcs)

    # func_core.calculation_time_equivalence()

    # WRITE RESULTS
    # =============

    return 0


if __name__ == '__main__':
    main([r"C:\Users\IanFu\Desktop\aaaaa"])
    # main()



# -*- coding: utf-8 -*-
# AUTHOR: YAN FU
# VERSION: 0.1
# DATE: 09/01/2018

import numpy as np
import os, time, matplotlib, inspect
import matplotlib.pyplot as plt
from inspect import currentframe, getframeinfo  # for error handling, get current line number
from scipy.interpolate import interp1d


class Scatter2D(object):
    def __init__(self, fixed_data_points = 0, fixed_data_points_range = (0, 0)):
        # todo: add plot vertical and horizontal lines feature
        # todo: re-think the overall logic

        self._figure = plt.figure()
        self._axes = []
        self._lines = []
        self._vlines = []
        self._hlines = []
        self._texts = []
        self._texts_added = []
        self._format = None

        self.__data_points = fixed_data_points
        self.__data_points_range = fixed_data_points_range

    def self_delete(self):
        self._figure.clf()
        plt.close(self._figure)
        # self._figure = None
        # self._axes = None
        # self._lines = None
        # self._texts = None
        del self

    def plot(self, xyl1, xyl2=None):
        # create _axes
        self._axes.append(self._figure.add_subplot(111))
        self._axes.append(self._axes[0].twinx()) if xyl2 is not None else None

        # create _lines
        for i in tuple(xyl1):
            x, y, l = tuple(i)
            line = self._axes[0].plot(x, y, label=l)
            self._lines.append(line[0])
        if xyl2 is not None:
            for i in xyl2:
                x, y, l = tuple(i)
                line = self._axes[1].plot(x, y, label=l)
                self._lines.append(line[0])

    def plot2(self, x, y, label="", second_axis=False, alpha=1.):

        # DATA POINTS NORMALISATION
        # =========================

        # Check if set by user

        if self.__data_points > 0:

            # interpolate new coordinates

            f_x = interp1d(x, y)

            if self.__data_points_range == (0, 0):
                x_new = np.linspace(np.min(x), np.max(x), self.__data_points)
            else:
                x_new = np.linspace(self.__data_points_range[0], self.__data_points_range[1], self.__data_points)

            y_new = f_x(x_new)

        else:
            x_new, y_new = x, y

        # PLOT
        # ====

        if len(self._axes) == 0:
            self._axes.append(self._figure.add_subplot(111))

        if second_axis:
            self._axes.append(self._axes[0].twinx())
            line = self._axes[1].plot(x_new, y_new, label=label, alpha=alpha)
        else:
            line = self._axes[0].plot(x_new, y_new, label=label, alpha=alpha)

        self._lines.append(line[0])

    def plot_vertical_line(self, x, line_color="black", line_linewidth=1):
        l = self._axes[0].axvline(x=x, color=line_color, linewidth=line_linewidth)
        self._vlines.append(l)

    def plot_horizontal_line(self, y, line_color="black", line_linewidth=1):
        l = self._axes[0].axhline(y=y, color=line_color, linewidth=line_linewidth)
        self._hlines.append(l)

    def format(self, **kwargs):
        def map_dictionary(list_, dict_master):
            dict_new = dict()
            for key in list_:
                if key in dict_master:
                    dict_new[key] = dict_master[key]
            return dict_new

        dict_inputs_figure = map_dictionary(inspect.signature(self.format_figure).parameters, kwargs)
        dict_inputs_axes = map_dictionary(inspect.signature(self.format_axes).parameters, kwargs)
        dict_inputs_lines = map_dictionary(inspect.signature(self.format_lines).parameters, kwargs)
        dict_inputs_legend = map_dictionary(inspect.signature(self.format_legend).parameters, kwargs)

        # set format
        self.format_figure(**dict_inputs_figure)
        self.format_axes(**dict_inputs_axes)
        self.format_lines(**dict_inputs_lines)
        self.format_legend(**dict_inputs_legend)

        self._figure.tight_layout()

    def format_figure(self,
                      figure_size_width=8.,
                      figure_size_height=6.,
                      figure_size_scale=1.,
                      figure_title="",
                      figure_title_font_size=15.):

        self._figure.set_size_inches(w=figure_size_width * figure_size_scale, h=figure_size_height * figure_size_scale)
        self._figure.suptitle(figure_title, fontsize=figure_title_font_size)
        self._figure.set_facecolor((1 / 237., 1 / 237., 1 / 237., 1.0))

    def format_axes(self,
                    axis_label_x="",
                    axis_label_y1="",
                    axis_label_y2="",
                    axis_label_font_size=9.,
                    axis_tick_font_size=8.,
                    axis_lim_x=None,
                    axis_lim_y1=None,
                    axis_lim_y2=None,
                    axis_linewidth=1.,
                    axis_scientific_format_x=False,
                    axis_scientific_format_y1=False,
                    axis_scientific_format_y2=False,
                    axis_tick_width=.5,
                    axis_tick_length=2.5,
                    axis_xtick_major_loc=None,
                    axis_xtick_minor_loc=None,
                    axis_ytick_major_loc=None,
                    axis_ytick_minor_loc=None,
                    axis_grid_show=False,
                    axis_grid_linestyle = "--",
                    axis_grid_linewidth = 0.25,
                    axis_grid_linecolour = "black",):
        has_secondary = len(self._axes) > 1

        self._axes[0].set_xlim(axis_lim_x)
        self._axes[0].set_ylim(axis_lim_y1)
        self._axes[1].set_ylim(axis_lim_y2) if has_secondary else None
        self._axes[0].set_xlabel(axis_label_x, fontsize=axis_label_font_size)
        self._axes[0].set_ylabel(axis_label_y1, fontsize=axis_label_font_size)
        self._axes[1].set_ylabel(axis_label_y2, fontsize=axis_label_font_size) if has_secondary else None
        self._axes[0].get_xaxis().get_major_formatter().set_useOffset(axis_scientific_format_x)
        self._axes[0].get_yaxis().get_major_formatter().set_useOffset(axis_scientific_format_y1)
        self._axes[1].get_yaxis().get_major_formatter().set_useOffset(axis_scientific_format_y2) if has_secondary else None

        [i.set_linewidth(axis_linewidth) for i in self._axes[0].spines.values()]
        [i.set_linewidth(axis_linewidth) for i in self._axes[1].spines.values()] if has_secondary else None

        self._axes[0].tick_params(axis='both', which='major', labelsize=axis_tick_font_size, width=axis_tick_width, length=axis_tick_length, direction='in')
        self._axes[0].tick_params(axis='both', which='minor', labelsize=axis_tick_font_size, width=axis_tick_width, length=axis_tick_length, direction='in')
        self._axes[1].tick_params(axis='both', which='major', labelsize=axis_tick_font_size, width=axis_tick_width, length=axis_tick_length, direction='in') if has_secondary else None
        self._axes[1].tick_params(axis='both', which='minor', labelsize=axis_tick_font_size, width=axis_tick_width, length=axis_tick_length, direction='in') if has_secondary else None

        if axis_xtick_major_loc is not None:
            self._axes[0].set_xticks(axis_xtick_major_loc)
        if axis_xtick_minor_loc is not None:
            self._axes[0].set_xticks(axis_xtick_minor_loc, minor=True)

        if axis_ytick_major_loc is not None:
            self._axes[0].set_yticks(axis_ytick_major_loc)
        if axis_ytick_minor_loc is not None:
            self._axes[0].set_yticks(axis_xtick_minor_loc, minor=True)

        if axis_grid_show:
            self._axes[0].grid(axis_grid_show, linestyle=axis_grid_linestyle, linewidth=axis_grid_linewidth, color=axis_grid_linecolour)
        else:
            self._axes[0].grid(axis_grid_show)


        # tick_lines = self._axes[0].get_xticklines() + self._axes[0].get_yticklines()
        # [line.set_linewidth(3) for line in tick_lines]
        #
        # tick_labels = self._axes[0].get_xticklabels() + self._axes[0].get_yticklabels()
        # [label.set_fontsize("medium") for label in tick_labels]

    def format_lines(self,
                     marker_size=3,
                     mark_every=100,
                     marker_fill_style="none",
                     marker_edge_width=.5,
                     line_width=1.,
                     line_style="-",
                     line_colours=None,
                     line_alpha=0):

        if line_colours is None:
            c = [(80, 82, 199), (30, 206, 214), (179, 232, 35), (245, 198, 0), (255, 89, 87)]
            c = [(colour[0] / 255., colour[1] / 255., colour[2] / 255.) for colour in c] * 100
        else:
            c = line_colours * 500

        m = ['o', '^', 's', 'v', 'p', '*', 'D', 'd', '8', '1', 'h', '+', 'H'] * 40

        for i, line in enumerate(self._lines):
            line.set_marker(m[i])
            line.set_color(c[i])
            line.set_markersize(marker_size)
            line.set_markevery(mark_every)
            line.set_markeredgecolor(c[i])
            line.set_markeredgewidth(marker_edge_width)
            line.set_fillstyle(marker_fill_style)
            line.set_linestyle(line_style)
            line.set_linewidth(line_width)

    def format_legend(self,
                      legend_is_shown=True,
                      legend_loc=0,
                      legend_font_size=8,
                      legend_colour="black",
                      legend_alpha=1.0,
                      legend_is_fancybox=False,
                      legend_line_width=1.):

        line_labels = [l.get_label() for l in self._lines]
        legend = self._axes[len(self._axes) - 1].legend(
            self._lines,
            line_labels,
            loc=legend_loc,
            fancybox=legend_is_fancybox,
            prop={'size': legend_font_size}
            )
        legend.set_visible(legend_is_shown)
        legend.get_frame().set_alpha(legend_alpha)
        legend.get_frame().set_linewidth(legend_line_width)
        legend.get_frame().set_edgecolor(legend_colour)

        self._texts.append(legend)

    def add_lines(self, xyl, axis=0):
        for i in xyl:
            x, y, l = tuple(i)
            line = self._axes[axis].plot(x, y, label=l)
            self._lines.append(line[0])

    def update_legend(self, **kwargs):
        """
        refresh the legend to the existing recent plotted _lines.
        """
        self._texts[0].remove()
        self.format_legend(**kwargs)

    def update_line_format(self, line_name, **kwargs):
        lines_index = {}
        for i,v in enumerate(self._lines):
            lines_index.update({v.get_label(): i})
        i = lines_index[line_name] if line_name in lines_index else None

        if i is None:
            frame_info = getframeinfo(currentframe())
            print("ERROR: {}; LINE: {:d}; FILE: {}".format("Line name does not exist", frame_info.lineno, frame_info.filename))
            return None

        line = self._lines[i]

        line_style = line.get_linestyle() if 'line_style' not in kwargs else kwargs['line_style']
        line_width = line.get_linewidth() if 'line_width' not in kwargs else kwargs['line_width']
        color = line.get_color() if 'color' not in kwargs else kwargs['color']
        marker = line.get_marker() if 'marker' not in kwargs else kwargs['marker']
        marker_size = line.get_markersize() if 'marker_size' not in kwargs else kwargs['marker_size']
        mark_every = line.get_markevery() if 'mark_every' not in kwargs else kwargs['mark_every']
        marker_edge_color = line.get_markeredgecolor() if 'marker_edge_color' not in kwargs else kwargs['marker_edge_color']
        marker_edge_width = line.get_markeredgewidth() if 'marker_edge_width' not in kwargs else kwargs['marker_edge_width']
        marker_fill_style = line.get_fillstyle() if 'marker_fill_style' not in kwargs else kwargs['marker_fill_style']

        line.set_linestyle(line_style)
        line.set_linewidth(line_width)
        line.set_color(color)
        line.set_marker(marker)
        line.set_markersize(marker_size)
        line.set_markevery(mark_every)
        line.set_markeredgecolor(marker_edge_color)
        line.set_markeredgewidth(marker_edge_width)
        line.set_fillstyle(marker_fill_style)

    def add_text(self, x, y, s, va="center", ha="center", fontsize=6):
        text_ = self._axes[0].text(x=x, y=y, s=s, va=va, ha=ha, fontsize=6)
        self._texts_added.append(text_)
        # self.adjust_text()

    def update_line(self, line_name, x, y, label=None):
        pass

    def remove_line(self, line_name):
        pass

    # def adjust_text(self):
    #
    #     # adjust_text(self._texts_added, arrowprops=dict(arrowstyle="->", color='r', lw=0.5),
    #     #             autoalign='xy',
    #     #             # only_move={'points': 'y', 'text': 'y'},
    #     #             expand_points=(5, 5),
    #     #             force_points=0.1)
    #
    #     adjust_text(self._texts_added,
    #                 # x=None,
    #                 # y=None,
    #                 # add_objects=None,
    #                 # ax=None,
    #                 expand_text=(1.2, 2.2),
    #                 expand_points=(5.2, 5.2),
    #                 expand_objects=(1.2, 1.2),
    #                 expand_align=(0.9, 0.9),
    #                 autoalign='xy',
    #                 va='center',
    #                 ha='center',
    #                 force_text=0.5,
    #                 force_points=0.5,
    #                 force_objects=0.5,
    #                 lim=100,
    #                 precision=0,
    #                 only_move={},
    #                 text_from_text=True,
    #                 text_from_points=True,
    #                 save_steps=False,
    #                 save_prefix='',
    #                 save_format='png',
    #                 add_step_numbers=True,
    #                 draggable=True,
    #                 on_basemap=False,
    #                 arrowprops=dict(arrowstyle="->", color='b', lw=0.5),)

    def __save_figure(self, file_name="_figure", file_format=".pdf", name_prefix="", name_suffix="", dir_folder="", dpi=300):
        # WARNING: DEPRECIATED!!!
        time_suffix = False
        str_time = time.strftime("%m%d.%H%M%S")
        if name_suffix == "time":
            name_suffix = str_time
        if name_prefix == "time":
            name_prefix = str_time
        file_name = "".join([name_prefix, file_name, name_suffix])
        self._figure.tight_layout()
        file_name += file_format
        # self.adjust_text()
        file_name = os.path.join(dir_folder, file_name)
        self._figure.savefig("/".join([dir_folder, file_name]), bbox_inches='tight', dpi=dpi)

    def save_figure2(self, path_file, dpi=300):
        self._figure.tight_layout()
        # self.adjust_text()
        self._figure.savefig(path_file, bbox_inches='tight', dpi=dpi)

    def show(self):
        self._figure.show(warn=True)

    @property
    def figure(self):
        return self.figure

    @property
    def axes_primary(self):
        return self._axes[0]

    @property
    def axes_secondary(self):
        return self._axes[1]


if __name__ == "__main__":
    x = np.arange(-2*np.pi, 2*np.pi, 0.01)
    y_sin = np.sin(x)
    y_cos = np.cos(x)
    y_tan = np.tan(x)

    p = Scatter2D()
    p.plot2(x, y_sin, 'sin(x)')
    p.plot2(x, y_cos, 'cos(x)')
    p.plot2(x, y_tan, 'tan(x)')
    p.plot2(x+np.pi/2, y_sin, 'sin(x+0.5pi)', second_axis=True)

    # default format
    plt_format = {
        'figure_size_width': 8.,
        'figure_size_height': 6.,
        'figure_size_scale': 1.,
        'figure_title': "",
        'figure_title_font_size': 15.,
        'axis_label_x': "",
        'axis_label_y1': "",
        'axis_label_y2': "",
        'axis_label_font_size': 9.,
        'axis_tick_font_size': 8.,
        'axis_lim_x': None,
        'axis_lim_y1': None,
        'axis_lim_y2': None,
        'axis_linewidth': 1.,
        'axis_scientific_format_x': False,
        'axis_scientific_format_y1': False,
        'axis_scientific_format_y2': False,
        'axis_tick_width': .5,
        'axis_tick_length': 2.5,
        'axis_xtick_major_loc': None,
        'axis_xtick_minor_loc': None,
        'axis_ytick_major_loc': None,
        'axis_ytick_minor_loc': None,
        'axis_grid_show': True,
        'axis_grid_linestyle': "-",
        'axis_grid_linewidth': 0.5,
        'axis_grid_linecolour': "black",
        'marker_size': 3,
        'mark_every': 100,
        'marker_fill_style': "none",
        'marker_edge_width': .5,
        'line_width': 1.,
        'line_style': "-",
        'line_colours': None,
    }

    # re-define some of the values
    plt_format_ = {
        'axis_lim_x': (0., 2*np.pi),
        'axis_lim_y1': (-1., 1.),
        'axis_lim_y2': (-2., 2.),
    }

    # update format dict
    plt_format.update(plt_format_)

    p.format(**plt_format)
    # p.save_figure(r"C:\hello")
    figure_file_path = os.path.abspath(r'hello.png')
    figure_file_path = os.path.realpath(figure_file_path)
    p.save_figure2(figure_file_path)



import os
import time
import multiprocessing as mp
import numpy as np
import pandas as pd
from pickle import load as pload
from pickle import dump as pdump
from scipy.interpolate import interp1d
import seaborn as sns
import matplotlib.pyplot as plt

from sfeprapy.func.temperature_fires import standard_fire_iso834 as _fire_standard
from sfeprapy.func.temperature_fires import parametric_eurocode1 as _fire_param
from sfeprapy.func.tfm_alt import travelling_fire as _fire_travelling
from sfeprapy.time_equivalence_mc import calc_time_equiv_worker, mc_inputs_generator

sns.set_style("ticks", {'axes.grid': True})

# OUTPUT STRING FORMAT
# ====================

__strformat_1_1 = "{:25}{}"
__strformat_1_1_1 = "{:25}{:3}{}"

# GLOBAL SETTINGS
# ===============

# Global variables
# -------

__app_pref = {
    'problem id': None,
    'problem folder directory': None,
    'result p name': 'res.p',
    'result csv name': 'res.csv',
    'result png name': 'teq_all.png',
    'result fire csv name': 'fires.csv'
}

# __id = ""
# __dir_work = ""
__fn_output = "res.p"
__fn_output_numerical = "res.csv"
__fn_plot_te = "teq.png"
__fn_plot_te_all = "teq_all.png"

# __fn_plot_temp = "plot_temp.png"

__fn_selected_plot = "fires.png"
__fn_fires_numerical = "fires.csv"


# def init_global_variables(_id, dir_work):
#     global __id, __dir_work
#     __id = _id
#     __dir_work = dir_work


def step0_parse_input_files(dir_work):
    """

    :param dir_work:
    :return:
    """

    list_files = []
    for f in os.listdir(dir_work):
        if f.endswith('.txt'):
            list_files.append(os.path.join(dir_work, f))

    list_input_files = []
    for f_ in list_files:
        with open(f_, "r") as f__:
            l = f__.readline()
        if l.find("# MC INPUT FILE") > -1:
            list_input_files.append(f_)

    return list_input_files


def step1_inputs_maker(path_input_file):
    """

    :param path_input_file:
    :return:
    """

    file_name = os.path.basename(path_input_file)
    dir_work = os.path.dirname(path_input_file)
    __id = file_name.split(".")[0]

    fire = _fire_standard(np.arange(0, 3 * 60 * 60, 1), 273.15 + 20)
    inputs_extra = {"iso834_time": fire[0],
                    "iso834_temperature": fire[1], }

    df_input, dict_pref = mc_inputs_generator(dict_extra_variables_to_add=inputs_extra, dir_file=path_input_file)

    return df_input, dict_pref


def step2_calc(df_input, dict_pref, path_input_file, progress_print_interval=5):
    # LOCAL SETTINGS
    # ==============

    # To limit memory usage when multiprocessing is employed, a maximum number of tasks is defined for a single process.
    # Therefore a process can not preserve data over this limit.

    mp_maxtasksperchild = 1000

    # Load kwargs

    dict_input_kwargs = df_input.to_dict(orient="index")
    list_kwargs = []
    for key, val in dict_input_kwargs.items():
        val["index"] = key
        list_kwargs.append(val)

    # Load settings

    dict_settings = dict_pref
    n_proc = dict_settings["n_proc"]

    # Check number of processes are to be used

    n_proc = os.cpu_count() if int(n_proc) < 1 else int(n_proc)

    # SIMULATION START

    print(__strformat_1_1.format("Input file:", os.path.basename(path_input_file)))
    print(__strformat_1_1.format("Total simulations:", len(list_kwargs)))
    print(__strformat_1_1.format("Number of threads:", n_proc))

    time_simulation_start = time.perf_counter()
    m = mp.Manager()
    q = m.Queue()
    p = mp.Pool(n_proc, maxtasksperchild=mp_maxtasksperchild)
    jobs = p.map_async(calc_time_equiv_worker, [(kwargs, q) for kwargs in list_kwargs])
    count_total_simulations = len(list_kwargs)
    n_steps = 24  # length of the progress bar
    while progress_print_interval:
        if jobs.ready():
            time_simulation_consumed = time.perf_counter() - time_simulation_start
            print("{}{} {:.1f}s".format('█' * round(n_steps), '-' * round(0), time_simulation_consumed))
            break
        else:
            p_ = q.qsize() / count_total_simulations * n_steps
            print("{}{} {:03.1f}%".format('█' * int(round(p_)), '-' * int(n_steps - round(p_)), p_ / n_steps * 100),
                  end='\r')
            time.sleep(1)
    p.close()
    p.join()
    results = jobs.get()

    # format outputs

    results = np.array(results)

    df_output = pd.DataFrame({'TIME STEP [s]': results[:, 0],
                              'TIME START [s]': results[:, 1],
                              'TIME LIMITING []': results[:, 2],
                              'WINDOW HEIGHT [m]': results[:, 3],
                              'WINDOW WIDTH [m]': results[:, 4],
                              'WINDOW OPEN FRACTION []': results[:, 5],
                              'ROOM BREADTH [m]': results[:, 6],
                              'ROOM DEPTH [m]': results[:, 7],
                              'ROOM HEIGHT [m]': results[:, 8],
                              'ROOM WALL THERMAL INERTIA [J/m2s1/2K]': results[:, 9],
                              'FIRE LOAD DENSITY [MJ/m2]': results[:, 10],
                              'FIRE HRR DENSITY [MW/m2]': results[:, 11],
                              'FIRE SPREAD SPEED [m/s]': results[:, 12],
                              'FIRE DURATION [s]': results[:, 13],
                              'BEAM POSITION [m]': results[:, 14],
                              'BEAM RHO [kg/m3]': results[:, 15],
                              'BEAM C [-]': results[:, 16],
                              'BEAM CROSS-SECTION AREA [m2]': results[:, 17],
                              'BEAM FAILURE TEMPERATURE [C]': results[:, 18],
                              'PROTECTION K [W/m/K]': results[:, 19],
                              'PROTECTION RHO [kg/m3]': results[:, 20],
                              'PROTECTION C OBJECT []': results[:, 21],
                              'PROTECTION THICKNESS [m]': results[:, 22],
                              'PROTECTION PERIMETER [m]': results[:, 23],
                              'ISO834 TIME ARRAY [s]': results[:, 24],
                              'ISO834 TEMPERATURE ARRAY [K]': results[:, 25],
                              'MAX. NEAR FIELD TEMPERATURE [C]': results[:, 26],
                              'SEEK ITERATION LIMIT []': results[:, 27],
                              'SEEK PROTECTION THICKNESS UPPER BOUND [m]': results[:, 28],
                              'SEEK PROTECTION THICKNESS LOWER BOUND [m]': results[:, 29],
                              'SEEK BEAM FAILURE TEMPERATURE TOLERANCE [K]': results[:, 30],
                              'INDEX': results[:, 31],
                              'TIME EQUIVALENCE [s]': results[:, 32],
                              'SEEK STATUS [0:Fail, 1:Success]': results[:, 33],
                              'FIRE TYPE [0:P, 1:T]': results[:, 34],
                              'SOUGHT BEAM TEMPERATURE [K]': results[:, 35],
                              'SOUGHT BEAM PROTECTION THICKNESS [m]': results[:, 36],
                              'SOUGHT ITERATIONS []': results[:, 37],
                              'BEAM TEMPERATURE TO FIXED PROTECTION THICKNESS [K]': results[:, 38],
                              'FIRE TIME ARRAY [s]': results[:, 39],
                              'FIRE TEMPERATURE ARRAY [K]': results[:, 40],
                              'OPENING FACTOR [m0.5]': results[:, 41]
                              })

    df_output.set_index("INDEX", inplace=True)  # assign 'INDEX' column as DataFrame index

    df_output.sort_values('TIME EQUIVALENCE [s]', inplace=True)  # sort base on time equivalence

    path_results_file = os.path.join(
        os.path.dirname(path_input_file),
        "{} - {}".format(
            os.path.basename(path_input_file).split('.')[0],
            __fn_output
        )
    )
    pdump(df_output, open(path_results_file, "wb"))

    return df_output


def plot_dist(id_, df_input, path_input_file, headers):

    # headers = ['window_open_fraction', 'fire_load_density', 'fire_spread_speed', 'beam_position', 'temperature_max_near_field']
    # headers = ['WINDOW OPEN FRACTION [%]', 'FIRE LOAD DENSITY [MJ/m2]', 'FIRE SPREAD SPEED [m/s]', 'BEAM POSITION [m]', 'MAX. NEAR FIELD TEMPERATURE [C]']

    names = {'WINDOW OPEN FRACTION []': 'Ao',
             'FIRE LOAD DENSITY [MJ/m2]': 'qfd',
             'FIRE SPREAD SPEED [m/s]': 'spread',
             'BEAM POSITION [m]': 'beam_loc',
             'MAX. NEAR FIELD TEMPERATURE [C]': 'nft'}

    fig, ax = plt.subplots(figsize=(3.94, 2.76))  # (3.94, 2.76) for large and (2.5, 2) for small figure size

    for k, v in names.items():
        x = np.array(df_input[k].values, float)

        if k == 'MAX. NEAR FIELD TEMPERATURE [C]':
            x = x[x < 1200]

        sns.distplot(x, kde=False, rug=True, bins=50, ax=ax, norm_hist=True)

        # Normal plot parameters
        ax.set_ylabel('PDF')
        ax.set_xlabel('x')

        # Small simple plot parameters
        # ax.set_ylabel('')
        # ax.set_yticklabels([])
        # ax.set_xlabel('')
        plt.tight_layout()
        plt.savefig(os.path.join(os.path.dirname(path_input_file), '{} - dist - {}.png'.format(id_, v)), ppi=300, transparent=True)
        plt.cla()

    plt.clf()


def select_fires_teq(df_output, path_input_file):

    dict_fires = dict()

    # for i,v in enumerate(list_index_selected_fires) :
    args = None
    for i, v in df_output.iterrows():
        # get input arguments
        # args = df_input.loc[i].to_dict()
        args = v.to_dict()

        # get fire type
        # fire_type = int(df_output.loc[i]["FIRE TYPE [0:P, 1:T]"])
        fire_type = int(v["FIRE TYPE [0:P, 1:T]"])

        if fire_type == 0:  # parametric fire
            w, l, h = args["ROOM BREADTH [m]"], args["ROOM DEPTH [m]"], args["ROOM HEIGHT [m]"]
            inputs_parametric_fire = {
                "A_t": 2 * (w * l + w * h + h * l),
                "A_f": w * l,
                "A_v": args["WINDOW HEIGHT [m]"] * args["WINDOW WIDTH [m]"] * args["WINDOW OPEN FRACTION []"],
                "h_eq": args["WINDOW HEIGHT [m]"],
                "q_fd": args["FIRE LOAD DENSITY [MJ/m2]"] * 1e6,
                "lambda_": args["ROOM WALL THERMAL INERTIA [J/m2s1/2K]"] ** 2,  # thermal inertia is used instead of k rho c.
                "rho": 1,  # see comment for lambda_
                "c": 1,  # see comment for lambda_
                "t_lim": args["TIME LIMITING []"],
                "time_end": args["FIRE DURATION [s]"],
                "time_step": args["TIME STEP [s]"],
                "time_start": args["TIME START [s]"],
                # "time_padding": (0, 0),
                "temperature_initial": 20 + 273.15,
            }
            tsec, temps = _fire_param(**inputs_parametric_fire)
        elif fire_type == 1:  # travelling fire
            inputs_travelling_fire = {
                "fire_load_density_MJm2": args["FIRE LOAD DENSITY [MJ/m2]"],
                "heat_release_rate_density_MWm2": args["FIRE HRR DENSITY [MW/m2]"],
                "length_compartment_m": args["ROOM DEPTH [m]"],
                "width_compartment_m": args["ROOM BREADTH [m]"],
                "fire_spread_rate_ms": args["FIRE SPREAD SPEED [m/s]"],
                "height_fuel_to_element_m": args["ROOM HEIGHT [m]"],
                "length_element_to_fire_origin_m": args["BEAM POSITION [m]"],
                "time_start_s": args["TIME START [s]"],
                "time_end_s": args["FIRE DURATION [s]"],
                "time_interval_s": args["TIME STEP [s]"],
                "nft_max_C": args["MAX. NEAR FIELD TEMPERATURE [C]"],
                "win_width_m": args["WINDOW WIDTH [m]"],
                "win_height_m": args["WINDOW HEIGHT [m]"],
                "open_fract": args["WINDOW OPEN FRACTION []"]
            }
            tsec, temps, hrr, r = _fire_travelling(**inputs_travelling_fire)
            temps += 273.15
        else:
            temps = 0
            print("FIRE TYPE UNKOWN.")

        dict_fires['FIRE {}'.format(str(i))] = temps - 273.15
        # plt.plot2(tsec/60., temps-273.15, alpha=.6)

    dict_fires["TIME [min]"] = np.arange(args["TIME START [s]"], args["FIRE DURATION [s]"], args["TIME STEP [s]"]) / 60.

    df_fires = pd.DataFrame(dict_fires)
    # list_names = ["TIME [min]"] + list_fire_name
    # df_fires = df_fires[list_names]

    # Save graphical plot to a .png file
    # ----------------------------------

    # Save numerical data to a .csv file
    # ----------------------------------
    file_name = "{} - {}".format(os.path.basename(path_input_file).split('.')[0], __fn_fires_numerical)
    df_fires.to_csv(os.path.join(os.path.dirname(path_input_file), file_name))


def step3_calc_post(list_path_input_file, list_pref, list_output):

    # SAVE NUMERICAL VALUES IN A .CSV FILE
    # ====================================

    for i, output_ in enumerate(list_output):
        # Obtain variable of individual simulation case
        # ---------------------------------------------

        path_input_file = list_path_input_file[i]

        # Obtain some useful strings / directories / variables
        # ----------------------------------------------------

        id_ = os.path.basename(path_input_file).split('.')[0]
        dir_work_ = os.path.dirname(path_input_file)

        # update global variables for file name and word path directory
        # init_global_variables(id_, dir_work_)

        # Save selected fires
        select_fires_teq(df_output=output_, path_input_file=path_input_file)

        # Save numerical values (inputs and outputs, i.e. everything)
        # -----------------------------------------------------------

        fn_ = os.path.join(dir_work_, " - ".join([id_, __fn_output_numerical]))

        output_.to_csv(fn_)

        # Plot distribution
        # -----------------
        if list_pref[i]['simulations'] > 2:
            plot_dist(id_, output_, path_input_file=path_input_file, headers=[])

        # Plot and save time equivalence for individual output
        # ----------------------------------------------------

        # obtain x and y values for plot
        x = np.sort(output_["TIME EQUIVALENCE [s]"].values) / 60.
        y = np.arange(1, len(x) + 1) / len(x)

    # PLOT TIME EQUIVALENCE FOR ALL OUTPUT
    # ====================================

    # do not plot if 'deterministic'

    line_vertical, line_horizontal = 0, 0
    teq_fig, teq_ax = plt.subplots(figsize=(3.94, 2.76))
    teq_ax.set_xlim([0, 120])
    for i, output_ in enumerate(list_output):

        # output_ = output_[output_['SEEK STATUS [0:Fail, 1:Success]'] == 1]

        # Obtain variable of individual simulation case
        # ---------------------------------------------

        path_input_file = list_path_input_file[i]
        pref_ = list_pref[i]

        # Obtain some useful strings / directories / variables
        # ----------------------------------------------------

        id_ = os.path.basename(path_input_file).split('.')[0]
        dir_work_ = os.path.dirname(path_input_file)

        # Check 'deterministic'
        # ----------------------
        if pref_['simulations'] <= 2:
            continue

        # obtain x and y values for plot

        x = np.sort(output_["TIME EQUIVALENCE [s]"].values) / 60.  # to minutes
        y = np.arange(1, len(x) + 1) / len(x)

        # plot the x, y

        plt.plot(x, y, label=id_)
        teq_ax.set_ylabel('Fractile')
        teq_ax.set_xlabel('Time [min]')
        teq_ax.set_xticks(ticks=np.arange(0, 180.1, 30))
        # plt.plot2(x, y, id_)
        # plt.format(**plt_format_)

        # plot horizontal and vertical lines (i.e. fractile and corresponding time euiqvalence period)

        if 0 < pref_["reliability_target"] < 1:
            line_horizontal = pref_["reliability_target"]
        elif pref_["reliability_target"] > 1:
            line_horizontal = 1 - 64.8 / pref_["reliability_target"] ** 2
        else:
            line_horizontal = 0

        if line_horizontal > 0:
            f_interp = interp1d(y, x)
            line_vertical = np.max((float(f_interp(line_horizontal)), line_vertical))

    if line_horizontal > 0:
        teq_ax.axhline(y=line_horizontal, c='grey')
        teq_ax.axvline(x=line_vertical, c='grey')
        teq_ax.text(
            x=line_vertical,
            y=teq_ax.get_ylim()[1],
            s="{:.0f}".format(line_vertical),
            va="bottom",
            ha="center",
            fontsize=9)

    teq_ax.legend().set_visible(True)
    teq_ax.legend(prop={'size': 7})
    plt.tight_layout()
    plt.savefig(
        os.path.join(os.path.dirname(list_path_input_file[0]), "{} - {}".format(os.path.basename(os.path.dirname(list_path_input_file[0])), __fn_plot_te)),
        bbox_inches='tight', dpi=300, transparent=True
    )

    plt.clf()
    plt.close()

__run_count = 0


# DEPRECIATED 20181114. MC portal migrated to /sfeprapy/mc.__main__
# def run(project_full_paths=list()):
#     _strfmt_1_1 = "{:25}{}"
#     __fn_output = "res.p"
#
#     global __run_count
#     __run_count += 1
#
#     if len(project_full_paths) == 0 or __run_count > 1:
#         from tkinter import filedialog, Tk, StringVar
#         root = Tk()
#         root.withdraw()
#         folder_path = StringVar()
#         while True:
#             file_name = filedialog.askdirectory(title='Select problem definitions folder')
#             if not file_name: break
#             folder_path.set(file_name)
#             project_full_paths.append(file_name)
#
#     # MAIN BODY
#     # =========
#
#     for project_full_path in project_full_paths:
#         list_path_input_file = step0_parse_input_files(dir_work=project_full_path)
#         list_path_input_file.sort()
#         list_input, list_output, list_pref, list_id = [], [], [], []
#         print(_strfmt_1_1.format("Work directory:", project_full_path))
#         for path_input_file in list_path_input_file:
#             id_ = os.path.basename(path_input_file).split(".")[0]
#
#             # Step 1: make a list of key word arguments for function inputs
#             # -------------------------------------------------------------
#
#             df_input, dict_pref = step1_inputs_maker(path_input_file)
#
#             ff = os.path.join(project_full_path, " - ".join([id_, __fn_output]))
#             if os.path.isfile(ff):
#                 df_output = pload(open(ff, "rb"))
#             else:
#
#                 # Step 2: perform main time equivalence calculation
#                 # -------------------------------------------------
#
#                 df_output = step2_calc(df_input, dict_pref, path_input_file)
#
#             list_input.append(df_input)
#             list_id.append(id_)
#             list_pref.append(dict_pref)
#             list_output.append(df_output)
#
#         # Step 3: select fire curves and output numerical and graphical files
#         # -------------------------------------------------------------------
#
#         step3_calc_post(list_path_input_file, list_pref, list_output)
#
#     input("Press Enter to finish")


if __name__ == '__main__':
    # run(['/Users/fuyans/Desktop/untitled'])
    # run()
    pass



# -*- coding: utf-8 -*-
# AUTHOR: YAN FU
# VERSION: 0.1
# DATE: 09/01/2018

# import numpy as np
# import os, time, matplotlib, inspect
# import matplotlib.pyplot as plt
# from inspect import currentframe, getframeinfo  # for error handling, get current line number
# from scipy.interpolate import interp1d


class Scatter2D(object):
    def __init__(self, fixed_data_points = 0, fixed_data_points_range = (0, 0)):
        pass

    def self_delete(self):
        pass

    def plot(self, xyl1, xyl2=None):
        pass

    def plot2(self, x, y, label="", second_axis=False, alpha=1.):
        pass

    def plot_vertical_line(self, x, line_color="black", line_linewidth=1):
        pass

    def plot_horizontal_line(self, y, line_color="black", line_linewidth=1):
        pass

    def format(self, **kwargs):
        pass

    def format_figure(self,
                      figure_size_width=8.,
                      figure_size_height=6.,
                      figure_size_scale=1.,
                      figure_title="",
                      figure_title_font_size=15.):

        pass

    def format_axes(self,
                    axis_label_x="",
                    axis_label_y1="",
                    axis_label_y2="",
                    axis_label_font_size=9.,
                    axis_tick_font_size=8.,
                    axis_lim_x=None,
                    axis_lim_y1=None,
                    axis_lim_y2=None,
                    axis_linewidth=1.,
                    axis_scientific_format_x=False,
                    axis_scientific_format_y1=False,
                    axis_scientific_format_y2=False,
                    axis_tick_width=.5,
                    axis_tick_length=2.5,
                    axis_xtick_major_loc=None,
                    axis_xtick_minor_loc=None,
                    axis_ytick_major_loc=None,
                    axis_ytick_minor_loc=None,
                    axis_grid_show=False,
                    axis_grid_linestyle = "--",
                    axis_grid_linewidth = 0.25,
                    axis_grid_linecolour = "black",):
        pass

    def format_lines(self,
                     marker_size=3,
                     mark_every=100,
                     marker_fill_style="none",
                     marker_edge_width=.5,
                     line_width=1.,
                     line_style="-",
                     line_colours=None,
                     line_alpha=0):

        pass

    def format_legend(self,
                      legend_is_shown=True,
                      legend_loc=0,
                      legend_font_size=8,
                      legend_colour="black",
                      legend_alpha=1.0,
                      legend_is_fancybox=False,
                      legend_line_width=1.):

        pass

    def add_lines(self, xyl, axis=0):
        pass

    def update_legend(self, **kwargs):
       pass

    def update_line_format(self, line_name, **kwargs):
        pass

    def add_text(self, x, y, s, va="center", ha="center", fontsize=6):
        pass

    def update_line(self, line_name, x, y, label=None):
        pass

    def remove_line(self, line_name):
        pass

    # def adjust_text(self):
    #
    #     # adjust_text(self._texts_added, arrowprops=dict(arrowstyle="->", color='r', lw=0.5),
    #     #             autoalign='xy',
    #     #             # only_move={'points': 'y', 'text': 'y'},
    #     #             expand_points=(5, 5),
    #     #             force_points=0.1)
    #
    #     adjust_text(self._texts_added,
    #                 # x=None,
    #                 # y=None,
    #                 # add_objects=None,
    #                 # ax=None,
    #                 expand_text=(1.2, 2.2),
    #                 expand_points=(5.2, 5.2),
    #                 expand_objects=(1.2, 1.2),
    #                 expand_align=(0.9, 0.9),
    #                 autoalign='xy',
    #                 va='center',
    #                 ha='center',
    #                 force_text=0.5,
    #                 force_points=0.5,
    #                 force_objects=0.5,
    #                 lim=100,
    #                 precision=0,
    #                 only_move={},
    #                 text_from_text=True,
    #                 text_from_points=True,
    #                 save_steps=False,
    #                 save_prefix='',
    #                 save_format='png',
    #                 add_step_numbers=True,
    #                 draggable=True,
    #                 on_basemap=False,
    #                 arrowprops=dict(arrowstyle="->", color='b', lw=0.5),)

    def save_figure2(self, path_file, dpi=300):
        pass

    def show(self):
        pass

    @property
    def figure(self):
        return None

    @property
    def axes_primary(self):
        return None

    @property
    def axes_secondary(self):
        return None


if __name__ == "__main__":
    # x = np.arange(-2*np.pi, 2*np.pi, 0.01)
    # y_sin = np.sin(x)
    # y_cos = np.cos(x)
    # y_tan = np.tan(x)
    #
    # p = Scatter2D()
    # p.plot2(x, y_sin, 'sin(x)')
    # p.plot2(x, y_cos, 'cos(x)')
    # p.plot2(x, y_tan, 'tan(x)')
    # p.plot2(x+np.pi/2, y_sin, 'sin(x+0.5pi)', second_axis=True)
    #
    # # default format
    # plt_format = {
    #     'figure_size_width': 8.,
    #     'figure_size_height': 6.,
    #     'figure_size_scale': 1.,
    #     'figure_title': "",
    #     'figure_title_font_size': 15.,
    #     'axis_label_x': "",
    #     'axis_label_y1': "",
    #     'axis_label_y2': "",
    #     'axis_label_font_size': 9.,
    #     'axis_tick_font_size': 8.,
    #     'axis_lim_x': None,
    #     'axis_lim_y1': None,
    #     'axis_lim_y2': None,
    #     'axis_linewidth': 1.,
    #     'axis_scientific_format_x': False,
    #     'axis_scientific_format_y1': False,
    #     'axis_scientific_format_y2': False,
    #     'axis_tick_width': .5,
    #     'axis_tick_length': 2.5,
    #     'axis_xtick_major_loc': None,
    #     'axis_xtick_minor_loc': None,
    #     'axis_ytick_major_loc': None,
    #     'axis_ytick_minor_loc': None,
    #     'axis_grid_show': True,
    #     'axis_grid_linestyle': "-",
    #     'axis_grid_linewidth': 0.5,
    #     'axis_grid_linecolour': "black",
    #     'marker_size': 3,
    #     'mark_every': 100,
    #     'marker_fill_style': "none",
    #     'marker_edge_width': .5,
    #     'line_width': 1.,
    #     'line_style': "-",
    #     'line_colours': None,
    # }
    #
    # # re-define some of the values
    # plt_format_ = {
    #     'axis_lim_x': (0., 2*np.pi),
    #     'axis_lim_y1': (-1., 1.),
    #     'axis_lim_y2': (-2., 2.),
    # }
    #
    # # update format dict
    # plt_format.update(plt_format_)
    #
    # p.format(**plt_format)
    # # p.save_figure(r"C:\hello")
    # figure_file_path = os.path.abspath(r'hello.png')
    # figure_file_path = os.path.realpath(figure_file_path)
    # p.save_figure2(figure_file_path)
    pass



# -*- coding: utf-8 -*-
import copy
import numpy as np
from scipy import stats
from scipy.interpolate import interp1d
from scipy.stats.distributions import norm, gumbel_r
from pandas import DataFrame as df

from sfeprapy.func.tfm_alt import travelling_fire as _fire_travelling
from sfeprapy.dat.steel_carbon import Thermal
from sfeprapy.func.temperature_steel_section import protected_steel_eurocode as _steel_temperature
from sfeprapy.func.temperature_fires import parametric_eurocode1 as _fire_param
from sfeprapy.func.kwargs_from_text import kwargs_from_text


def lognorm_parameters_true_to_inv(miu, sigma):
    """
    NAME: lognorm_parameters_true_to_inv
    VERSION: 0.0.1
    AUTHOR: Yan Fu, Ruben V. Coline
    DESCRIPTION:
    Converts the mean and standard deviation of distribution from x to ln(x).

    PARAMETERS:
    :param miu: True mean of the x
    :param sigma: True standard deviation of x
    :return: (miu, sigma) where miu and sigma are based on ln(x)

    USAGE:
    >>> print(lognorm_parameters_true_to_inv(0.2, 0.2))
    >>> (-1.9560115027140728, 0.8325546111576977)
    """
    cov = sigma / miu

    sigma_ln = np.sqrt(np.log(1 + cov ** 2))
    miu_ln = np.log(miu) - 1 / 2 * sigma_ln ** 2

    return miu_ln, sigma_ln


# if __name__ == '__main__':
#     print(lognorm_parameters_true_to_inv(0.2, 0.2))


def lognorm_trunc_ppf(a, b, n_rv, sigma, loc, scale, cdf_y=None):
    """
    NAME: lognorm_trunc_ppf
    VERSION: 0.0.1
    AUTHOR: Yan Fu
    DATE: 3 Aug 2018
    DESCRIPTION:
    Truncated log normal distribution cumulative function density. Truncate and normalise  a log normal distribution
    function for a given boundary (a, b).

    PARAMETERS:
    :param a: float, Lower boundary
    :param b: float, Upper boundary
    :param n_rv: integer, total number of
    :param sigma: float, standard deviation of log normal distribution (for ln(x))
    :param loc: float, location of log normal distribution
    :param scale: float, scale of the log normal distribution
    :param cdf_y: array (1 dimension) or None. A set of numbers represent cumulative probability. If None the function
    will return the sampled values
    :return: array (1 dimension), set of numbers represent sampled values of truncated log normal distribution inline
    with 'cfd_y'

    USAGE:
    >>> import numpy as np
    >>> a = 0                   # lower boundary
    >>> b = 1                   # upper boundary
    >>> n_rv = 5                # number of random variables
    >>> sigma = 0.2             # standard deviation
    >>> miu = 0.2               # mean

    # Convert true mean and sigma to ln(x) based mean and sigma
    >>> miu, sigma = lognorm_parameters_true_to_inv(miu, sigma)

    >>> loc = np.exp(miu)       # distribution mean is 0.2 (i.e. loc = np.exp(miu))
    >>> scale = 1               # default
    >>> result = lognorm_trunc_ppf(a, b, n_rv, sigma, loc, scale)
    >>> print(result)
    [0.14142136 0.49653783 0.65783987 0.81969479 1.        ]
    """

    # Generate a linear spaced array inline with lower and upper boundary of log normal cumulative probability density.
    sampled_cfd = np.linspace(
        stats.lognorm.cdf(x=a, s=sigma, loc=loc, scale=scale),
        stats.lognorm.cdf(x=b, s=sigma, loc=loc, scale=scale),
        n_rv
    )

    # Sample log normal distribution
    sampled = stats.lognorm.ppf(q=sampled_cfd, s=sigma, loc=loc, scale=scale)

    # Work out cumulative probability function from 'sampled', output in forms of x y.
    # Interpolate x and y are processed to be capable to cope with two extreme values. y[0] (cumulative probability,
    # initial boundary) is manually set to 0.
    x = np.linspace(a, b, int(n_rv), endpoint=False)
    x += (x[1] - x[0]) / 2
    x[-1] -= (x[1] - x[0]) / 2
    x = np.append([0], x)
    y = np.array([np.sum(sampled <= i) for i in x]) / len(sampled)
    y[0] = 0

    # Interpolate
    f = interp1d(y, x, bounds_error=False, fill_value=(np.min(y), np.max(y)))

    if cdf_y is None:
        return sampled
    else:
        return f(cdf_y)


def gumbel_parameter_converter(miu, sigma):
    """
    NAME: gumbel_parameter_converter
    VERSION: 0.0.1
    AUTHOR: Yan Fu, Ruben V. Coline
    DATE: 27 Sept 2018
    DESCRIPTION:
    This function is used in conjunction of the scipy.stats.gumbel distribution, converts mean and standard deviation
    (sigma) of samples x to location and scale which is used in scipy.stats.gumbel_r function.

    :param miu: mean value of samples x
    :param sigma: standard deviation of samples x
    :return: location and scale

    EXAMPLE:

    """

    # parameters Gumbel W&S
    alpha = 1.282 / sigma
    u = miu - 0.5772 / alpha

    # parameters Gumbel scipy
    scale = 1 / alpha
    loc = u

    return loc, scale


def gumbel_r_trunc_ppf(a, b, n_rv, loc, scale, cdf_y=None):
    """
    NAME: gumbel_r_trunc_ppf
    VERSION: 0.0.1
    AUTHOR: Yan Fu
    DATE: 27 Sept 2018
    DESCRIPTION:
    Produces evenly sampled random variables based on gumbel distribution (tail to the x+ direction, i.e. median greater
    than mean). Truncation is possible via variables 'a' and 'b'. i.e. inversed cumulative density function f(x), x will
    be sampled in linear space ranging from 'a' to 'b'. Then f(x) is returned. Additionally, if x is defined 'cdf_y'
    then f(cdf_y) is returned.

    PARAMETERS:
    :param a: lower bound of truncation
    :param b: upper bound of truncation
    :param n_rv: number of random variables to be sampled, equal to the length of the returned array
    :param loc: location of the distribution
    :param scale: scale of the distribution
    :param cdf_y: array ranging with range (0, 1)
    :return sampled: sampled random variables

    USAGE:
    >>> import numpy as np
    >>> import matplotlib.pyplot as plt
    >>> loc, scale = gumbel_parameter_converter(600, 180)
    >>> rv = gumbel_r_trunc_ppf(0, 1800, 10, loc, scale)
    >>> print(np.round(rv, 0))
    [   0.  408.  462.  506.  548.  594.  646.  713.  819. 1800.]

    """

    # Generate a linear spaced array inline with lower and upper boundary of log normal cumulative probability density.
    sampled_cfd = np.linspace(
        stats.gumbel_r.cdf(x=a, loc=loc, scale=scale),
        stats.gumbel_r.cdf(x=b, loc=loc, scale=scale),
        n_rv
    )

    # Following three lines are used to check the validity of the distribution
    # print("511 (0.80): {:.4f}".format(stats.gumbel_r.cdf(x=510, loc=loc, scale=scale)))
    # print("584 (0.90): {:.4f}".format(stats.gumbel_r.cdf(x=584, loc=loc, scale=scale)))
    # print("655 (0.95): {:.4f}".format(stats.gumbel_r.cdf(x=655, loc=loc, scale=scale)))

    # Sample log normal distribution
    sampled = stats.gumbel_r.ppf(q=sampled_cfd, loc=loc, scale=scale)

    # Work out cumulative probability function from 'sampled', output in forms of x y.
    # Interpolate x and y are processed to be capable to cope with two extreme values. y[0] (cumulative probability,
    # initial boundary) is manually set to 0.
    x = np.linspace(a, b, int(n_rv) + 1, endpoint=False)
    x += (x[1] - x[0]) / 2
    x = x[0:-2]
    # x[-1] -= (x[1] - x[0]) / 2
    # x = np.append([0], x)
    y = np.array([np.sum(sampled <= i) for i in x]) / len(sampled)
    # y[0] = 0

    # Interpolate
    f = interp1d(y, x, bounds_error=False, fill_value=(np.min(y), np.max(y)))

    if cdf_y is None:
        return sampled
    else:
        return f(cdf_y)


if __name__ == '__main__':
    import numpy as np
    import seaborn as sns
    import matplotlib.pyplot as plt
    loc, scale = gumbel_parameter_converter(600, 180)
    rv = gumbel_r_trunc_ppf(0, 1800, 10, loc, scale)
    print(np.round(rv, 0))
    sns.distplot(rv)
    plt.show()


def latin_hypercube_sampling(num_samples, num_arguments=1, sample_lbound=0, sample_ubound=1):
    """
    NAME: latin_hypercube_sampling
    AUTHOR: Yan Fu
    VERSION: 0.1
    DATE: 3 Aug 2018
    DESCRIPTION:
    Latin Hypercube Sampling, generates an nxm array where m equal to 'num_arguments' and n equal to 'num_samples'.
    Current version only adopts 'centered' sampling mode (each sampled cell value is centered).

    PARAMETERS:
    :param num_samples: Number of samples (i.e. rows)
    :param num_arguments: Number of arguments (i.e. columns)
    :param sample_lbound: Lower sampling boundary
    :param sample_ubound: Upper sampling boundary
    :return: An array with shape (num_samples, num_arguments)

    EXAMPLE:
    >>> result = latin_hypercube_sampling(num_samples=10, num_arguments=3, sample_lbound=0, sample_ubound=0.001)
    This example yields an array with shape of (100, 3), with each column filled 100 linear spaced numbers (shuffled)
    from 1 to 0.001 (i.e. shuffled [0.05, 0.15, 0.25, 0.35, 0.45, 0.55, 0.65, 0.75, 0.85, 0.95]).
    An example output:
    result = [
        [0.85 0.95 0.55]
        [0.55 0.45 0.05]
        [0.25 0.25 0.15]
        [0.35 0.85 0.75]
        [0.45 0.65 0.85]
        [0.75 0.05 0.65]
        [0.05 0.35 0.45]
        [0.15 0.15 0.35]
        [0.65 0.55 0.25]
        [0.95 0.75 0.95]
    ]

    """

    if sample_lbound > sample_ubound:
        sample_ubound += sample_lbound
        sample_lbound = sample_ubound - sample_lbound
        sample_ubound = sample_ubound - sample_lbound

    # Generate sorted integers with correct shape
    mat_random_num = np.linspace(sample_lbound, sample_ubound, num_samples + 1, dtype=float)
    mat_random_num += (mat_random_num[1] - mat_random_num[0]) * 0.5
    mat_random_num = mat_random_num[0:-1]
    mat_random_num = np.reshape(mat_random_num, (len(mat_random_num), 1))
    mat_random_nums = mat_random_num * np.ones((1, num_arguments))

    # np.random.shuffle(mat_random_nums)

    for i in range(np.shape(mat_random_nums)[1]):
        np.random.shuffle(mat_random_nums[:, i])

    if num_arguments == 1:
        mat_random_nums = mat_random_nums.flatten()

    return mat_random_nums


def calc_time_equiv_worker(arg):
    kwargs, q = arg
    result = calc_time_equivalence(**kwargs)
    q.put("index: {}".format(kwargs["index"]))
    return result


def calc_time_equivalence(
        time_step,
        time_limiting,
        window_height,
        window_width,
        window_open_fraction,
        room_breadth,
        room_depth,
        room_height,
        room_wall_thermal_inertia,
        fire_load_density,
        fire_hrr_density,
        fire_spread_speed,
        fire_duration,
        beam_position,
        beam_rho,
        beam_c,
        beam_cross_section_area,
        beam_temperature_goal,
        beam_loc_z,
        protection_k,
        protection_rho,
        protection_c,
        protection_thickness,
        protection_protected_perimeter,
        iso834_time,
        iso834_temperature,
        index,
        time_start=0,
        parametric_fire_enforce=False,
        nft_ubound=1200,
        seek_ubound_iter=20,
        seek_ubound=0.1,
        seek_lbound=0.0001,
        seek_tol_y=1.,
        return_mode=0,
        **kwargs
):
    """
    NAME: calc_time_equivalence
    VERSION: 0.0.1
    AUTHOR: Ian Fu
    DATE: 1 Oct 2018
    DESCRIPTION:
    Calculates equivalent time exposure for a protected steel element member in more realistic fire environment
    opposing to the standard fire curve ISO 834.

    PARAMETERS:
    :param time_step: [s], time step used for numerical calculation
    :param time_start: [s], simulation starting time
    :param time_limiting: [-], PARAMETRIC FIRE, see parametric fire function for details
    :param window_height: [m], weighted window opening height
    :param window_width: [m], total window opening width
    :param window_open_fraction: [-], a factor is multiplied with the given total window opening area
    :param room_breadth: [m], room breadth (shorter direction of the floor plan)
    :param room_depth: [m], room depth (longer direction of the floor plan)
    :param room_height: [m], room height from floor to soffit (structural), disregard any non fire resisting floors
    :param room_wall_thermal_inertia: [J m-2 K-1 s-1/2], thermal inertia of room lining material
    :param fire_load_density: [MJ m-2], fire load per unit area
    :param fire_hrr_density: [MW m-2], fire maximum release rate per unit area
    :param fire_spread_speed: [m s-1], TRAVELLING FIRE, fire spread speed
    :param fire_duration: [s], simulation time
    :param beam_position: [s], beam location
    :param beam_rho: [kg m-3], density of the steel beam element
    :param beam_c: [?], specific heat of the steel element
    :param beam_cross_section_area: [m2], the steel beam element cross section area
    :param beam_temperature_goal: [K], steel beam element expected failure temperature
    :param protection_k: steel beam element protection material thermal conductivity
    :param protection_rho: steel beam element protection material density
    :param protection_c: steel beam element protection material specific heat
    :param protection_thickness: steel beam element protection material thickness
    :param protection_protected_perimeter: [m], steel beam element protection material perimeter
    :param iso834_time: [s], the time (array) component of ISO 834 fire curve
    :param iso834_temperature: [K], the temperature (array) component of ISO 834 fire curve
    :param nft_ubound: [K], TRAVELLING FIRE, maximum temperature of near field temperature
    :param seek_ubound_iter: Maximum allowable iteration counts for seeking solution for time equivalence
    :param seek_ubound: [m], protection layer thickness upper bound initial condition for solving time equivalence
    :param seek_lbound: [m], protection layer thickness lower bound initial condition for solving time equivalence
    :param seek_tol_y: [K], tolerance for solving time equivalence
    :param index: will be returned for indicating the index of the current iteration (was used for multiprocessing)
    :param return_mode: 0 - minimal for teq; 1 - all variables; and 2 - all variables packed in a dict
    :param kwargs: will be discarded
    :return:

    EXAMPLE:
    """

    #   Check on applicable fire curve
    window_area = window_height * window_width * window_open_fraction
    room_floor_area = room_breadth * room_depth
    room_area = (2 * room_floor_area) + ((room_breadth + room_depth) * 2 * room_height)

    #   Opening factor - is it within EC limits?
    opening_factor = window_area * np.sqrt(window_height) / room_area

    #   Spread speed - Does the fire spread to involve the full compartment?
    sp_time = max([room_depth, room_breadth]) / fire_spread_speed
    burnout_m2 = max([fire_load_density / fire_hrr_density, 900.])

    if parametric_fire_enforce:
        opening_factor = min(0.2, max(0.02, opening_factor))  # force opening factor fall in the boundary

        fire_time, fire_temp = _fire_param(**{"A_t": room_area,
                                              "A_f": room_floor_area,
                                              "A_v": window_area,
                                              "h_eq": window_height,
                                              "q_fd": fire_load_density * 1e6,
                                              "lambda_": room_wall_thermal_inertia ** 2,
                                              "rho": 1,
                                              "c": 1,
                                              "t_lim": time_limiting,
                                              "time_end": fire_duration,
                                              "time_step": time_step,
                                              "time_start": time_start,
                                              # "time_padding": (0, 0),
                                              "temperature_initial": 20 + 273.15, })

        fire_type = 0  # parametric fire

    elif sp_time < burnout_m2 and 0.02 < opening_factor <= 0.2:  # If fire spreads throughout compartment and ventilation is within EC limits = Parametric fire
        fire_time, fire_temp = _fire_param(**{"A_t": room_area,
                                              "A_f": room_floor_area,
                                              "A_v": window_area,
                                              "h_eq": window_height,
                                              "q_fd": fire_load_density * 1e6,
                                              "lambda_": room_wall_thermal_inertia ** 2,
                                              "rho": 1,
                                              "c": 1,
                                              "t_lim": time_limiting,
                                              "time_end": fire_duration,
                                              "time_step": time_step,
                                              "time_start": time_start,
                                              # "time_padding": (0, 0),
                                              "temperature_initial": 20 + 273.15, })

        fire_type = 0  # parametric fire

    else:  # Otherwise, it is a travelling fire
        # Get travelling fire curve
        fire_time, fire_temp, heat_release, distance_to_element = _fire_travelling(
            fire_load_density,
            fire_hrr_density,
            room_depth,
            room_breadth,
            fire_spread_speed,
            beam_loc_z,
            beam_position,
            time_start,
            fire_duration,
            time_step,
            nft_ubound,
            window_width,
            window_height,
            window_open_fraction,
        )
        fire_temp += 273.15
        fire_type = 1  # travelling fire

    # Solve heat transfer using EC3 correlations
    # SI UNITS FOR INPUTS!
    inputs_steel_heat_transfer = {"time": fire_time,
                                  "temperature_ambient": fire_temp,
                                  "rho_steel": beam_rho,
                                  "c_steel_T": beam_c,
                                  "area_steel_section": beam_cross_section_area,
                                  "k_protection": protection_k,
                                  "rho_protection": protection_rho,
                                  "c_protection": protection_c,
                                  "thickness_protection": protection_thickness,
                                  "perimeter_protected": protection_protected_perimeter,
                                  "is_terminate_peak": True}

    # Find maximum steel temperature for the static protection layer thickness
    if protection_thickness > 0:
        temperature_steel_ubound = np.max(_steel_temperature(**inputs_steel_heat_transfer)[1])
    else:
        temperature_steel_ubound = -1

    # ============================================
    # GOAL SEEK TO MATCH STEEL FAILURE TEMPERATURE
    # ============================================

    # MATCH PEAK STEEL TEMPERATURE BY ADJUSTING PROTECTION LAYER THICKNESS

    seek_count_iter = 0  # count how many iterations for  the seeking process
    seek_status = False  # flag used to indicate when the seeking is successful

    # Default values
    fire_resistance_equivalence = -1
    sought_temperature_steel_ubound = -1
    sought_protection_thickness = -1

    # import matplotlib.pyplot as plt
    # fig, ax = plt.subplots()
    # ax.plot(fire_time, fire_temp)

    if beam_temperature_goal > 0:  # check seeking temperature, opt out if less than 0
        # Seeking process
        while seek_count_iter < seek_ubound_iter and seek_status is False:
            seek_count_iter += 1
            sought_protection_thickness = np.average([seek_ubound, seek_lbound])
            inputs_steel_heat_transfer["thickness_protection"] = sought_protection_thickness
            t_, T_, d_ = _steel_temperature(**inputs_steel_heat_transfer)
            sought_temperature_steel_ubound = np.max(T_)
            y_diff_seek = sought_temperature_steel_ubound - beam_temperature_goal
            if abs(y_diff_seek) <= seek_tol_y:
                seek_status = True
            elif sought_temperature_steel_ubound > beam_temperature_goal:  # steel too hot, increase protect thickness
                seek_lbound = sought_protection_thickness
            else:  # steel is too cold, increase intrumescent paint thickness
                seek_ubound = sought_protection_thickness

            # ax.plot(t_, T_)
            # plt.draw()
            # plt.show()

        # CALCULATE BEAM FIRE RESISTANCE PERIOD IN ISO 834
        # ======================================

        # Make steel time-temperature curve when exposed to the given ambient temperature, i.e. ISO 834.
        inputs_steel_heat_transfer["time"] = iso834_time
        inputs_steel_heat_transfer["temperature_ambient"] = iso834_temperature
        time_, temperature_steel, data_all = _steel_temperature(**inputs_steel_heat_transfer)

        # re-arrange time and steel temperature, which will be used later, to prevent interpolation boundary error. If
        # boundaries are breached, the corresponding min or max of time will be returned, i.e. 0 and 2 hrs
        time_ = np.concatenate((np.array([0]), time_, np.array([time_[-1]])))
        temperature_steel = np.concatenate((np.array([-1]), temperature_steel, np.array([1e12])))

        # CALCULATE EQUIVALENCE EXPOSURE TIME IN ISO 834 TO 620 C

        # perform interpolation for teq based on acutal steel temperature and iso steel temperature
        interp_ = interp1d(temperature_steel, time_, kind="linear", bounds_error=False, fill_value=-1)
        fire_resistance_equivalence = interp_(beam_temperature_goal)

    r_dict = {'room_wall_thermal_inertia': room_wall_thermal_inertia,
              'fire_load_density': fire_load_density,
              'fire_hrr_density': fire_hrr_density,
              'fire_spread_speed': fire_spread_speed,
              'fire_duration': fire_duration,
              'beam_position': beam_position,
              'beam_rho': beam_rho,
              'beam_c': beam_c,
              'beam_cross_section_area': beam_cross_section_area,
              'beam_temperature_goal': beam_temperature_goal,
              'protection_k': protection_k,
              'protection_rho': protection_rho,
              'protection_c': protection_c,
              'protection_thickness': protection_thickness,
              'protection_protected_perimeter': protection_protected_perimeter,
              'iso834_time': iso834_time,
              'iso834_temperature': iso834_temperature,
              'nft_ubound': nft_ubound,
              'seek_ubound_iter': seek_ubound_iter,
              'seek_ubound': seek_ubound,
              'seek_lbound': seek_lbound,
              'seek_tol_y': seek_tol_y,
              'index': index,
              'fire_resistance_equivalence': fire_resistance_equivalence,
              'seek_status': seek_status,
              'fire_type': fire_type,
              'sought_temperature_steel_ubound': sought_temperature_steel_ubound,
              'sought_protection_thickness': sought_protection_thickness,
              'seek_count_iter': seek_count_iter,
              'temperature_steel_ubound': temperature_steel_ubound,
              'fire_time': fire_time,
              'fire_temp': fire_temp}

    if return_mode == 0:  # made all objects equal to zero to reduce the size
        return time_step, time_start, time_limiting, window_height, window_width, window_open_fraction, room_breadth, room_depth, room_height, room_wall_thermal_inertia, fire_load_density, fire_hrr_density, fire_spread_speed, fire_duration, beam_position, beam_rho, 0, beam_cross_section_area, beam_temperature_goal, protection_k, protection_rho, protection_c, protection_thickness, protection_protected_perimeter, 0, 0, nft_ubound, seek_ubound_iter, seek_ubound, seek_lbound, seek_tol_y, index, fire_resistance_equivalence, seek_status, fire_type, sought_temperature_steel_ubound, sought_protection_thickness, seek_count_iter, temperature_steel_ubound, 0, 0, opening_factor
    elif return_mode == 1:  # full results, contains everything, including objects
        return time_step, time_start, time_limiting, window_height, window_width, window_open_fraction, room_breadth, room_depth, room_height, room_wall_thermal_inertia, fire_load_density, fire_hrr_density, fire_spread_speed, fire_duration, beam_position, beam_rho, beam_c, beam_cross_section_area, beam_temperature_goal, protection_k, protection_rho, protection_c, protection_thickness, protection_protected_perimeter, iso834_time, iso834_temperature, nft_ubound, seek_ubound_iter, seek_ubound, seek_lbound, seek_tol_y, index, fire_resistance_equivalence, seek_status, fire_type, sought_temperature_steel_ubound, sought_protection_thickness, seek_count_iter, temperature_steel_ubound, fire_time, fire_temp, opening_factor
    elif return_mode == 2:
        return r_dict


def mc_inputs_generator(dict_vars, dict_extra_variables_to_add=None):
    steel_prop = Thermal()

    # ------------------------------------------------------------------------------------------------------------------
    # Define the inputs from file
    # ------------------------------------------------------------------------------------------------------------------

    # dict_setting_vars = dict()

    # Read input variables from external text file
    # with open(str(dir_file), "r") as file_inputs:
    #     string_inputs = file_inputs.read()
    # dict_vars_0 = kwargs_from_text(string_inputs)
    if dict_extra_variables_to_add:
        dict_vars.update(dict_extra_variables_to_add)

    dict_vars["beam_c"] = steel_prop.c()

    # dict_vars_0_ is used for making the returned DataFrame, hence it passes necessary variables only
    # dict_vars_0_ = copy.copy(dict_vars_0)

    return mc_inputs_generator_core(**dict_vars)


def mc_inputs_generator_core(
        # Setting parameters
        simulations,
        n_proc,
        reliability_target,
        # select_fires_teq,
        # select_fires_teq_tol,

        # Simulation parameters
        room_breadth,
        room_depth,
        room_height,
        window_width,
        window_height,

        time_start,
        time_limiting,
        time_step,
        fire_duration,
        fire_hrr_density,
        room_wall_thermal_inertia,

        beam_cross_section_area,
        beam_rho,
        beam_c,
        beam_temperature_goal,
        protection_protected_perimeter,
        protection_thickness,
        protection_k,
        protection_rho,
        protection_c,

        qfd_std,
        qfd_mean,
        qfd_ubound,
        qfd_lbound,
        opening_fraction_lbound,
        opening_fraction_ubound,
        opening_fraction_std,
        opening_fraction_mean,
        beam_loc_ratio_lbound,
        beam_loc_ratio_ubound,
        com_eff_lbound,
        com_eff_ubound,
        spread_lbound,
        spread_ubound,
        nft_mean,

        # Derived
        iso834_time,
        iso834_temperature,

        # Control
        # is_deterministic,
):

    # linear_distribution = lambda min, max, prob: ((max - min) * prob) + min
    def linear_distribution(min_, max_, rv_):
        return ((max_ - min_) * rv_) + min_

    # CHECKS
    # ======

    # Fire duration has to be as long as travelling fire to travel through the entire floor
    fire_duration = np.max([fire_duration, room_depth / spread_lbound])

    # ------------------------------------------------------------------------------------------------------------------
    # Distribution variables
    # ------------------------------------------------------------------------------------------------------------------

    if simulations > 2:

        # lhs_mat = lhs(n=6, samples=simulations, criterion=dict_setting_vars["lhs_criterion"])
        lhs_mat = latin_hypercube_sampling(num_samples=simulations, num_arguments=6)

        # Near field standard deviation
        # nft_mean = nft_mean  # TFM near field temperature - Norm distribution - mean [C]
        std_nft = (1.939 - (np.log(nft_mean) * 0.266)) * nft_mean

        # Convert LHS probabilities to distribution invariants
        comb_lhs = linear_distribution(com_eff_lbound, com_eff_ubound, lhs_mat[:, 0])

        # Fuel load density
        # -----------------
        qfd_loc, qfd_scale = gumbel_parameter_converter(qfd_mean, qfd_std)
        qfd_lhs = gumbel_r_trunc_ppf(qfd_lbound, qfd_ubound, simulations, qfd_loc, qfd_scale) * comb_lhs
        np.random.shuffle(qfd_lhs)
        # DEPRECIATED 14 Aug 2018 - Following codes calculate gumbel parameters for qfd
        # qfd_scale = qfd_std * (6 ** 0.5) / np.pi
        # qfd_loc = qfd_mean - (0.57722 * qfd_scale)
        # qfd_dist = gumbel_r(loc=qfd_loc, scale=qfd_scale)
        # qfd_p_l, qfd_p_u = qfd_dist.cdf(qfd_lbound), qfd_dist.cdf(qfd_ubound)
        # qfd_lhs = gumbel_r(loc=qfd_loc, scale=qfd_scale).ppf(lhs_mat[:, 1]) * comb_lhs

        # Opening fraction factor (glazing fall-out fraction)
        # ---------------------------------------------------
        opening_fraction_mean_conv, opening_fraction_std_conv = lognorm_parameters_true_to_inv(opening_fraction_mean,
                                                                                               opening_fraction_std)
        glaz_lhs = 1 - lognorm_trunc_ppf(opening_fraction_lbound, opening_fraction_ubound, simulations,
                                         opening_fraction_std_conv, 0, np.exp(opening_fraction_mean_conv))
        np.random.shuffle(glaz_lhs)

        # Beam location
        # -------------
        beam_lhs = linear_distribution(beam_loc_ratio_lbound, beam_loc_ratio_ubound, lhs_mat[:, 3]) * room_depth

        # Fire spread speed (travelling fire)
        # -----------------------------------
        spread_lhs = linear_distribution(spread_lbound, spread_ubound, lhs_mat[:, 4])

        # Near field temperature (travelling fire)
        # ----------------------------------------
        nft_lhs = norm(loc=nft_mean, scale=std_nft).ppf(lhs_mat[:, 5])
        nft_lhs[nft_lhs > 1200] = 1200  # todo: reference?

    # ------------------------------------------------------------------------------------------------------------------
    # Create input kwargs for mc calculation
    # ------------------------------------------------------------------------------------------------------------------

    dict_input_kwargs_static = dict(
        room_breadth=room_breadth,
        room_depth=room_depth,
        room_height=room_height,
        window_width=window_width,
        window_height=window_height,

        time_start=time_start,
        time_limiting=time_limiting,
        time_step=time_step,
        fire_duration=fire_duration,
        fire_hrr_density=fire_hrr_density,
        room_wall_thermal_inertia=room_wall_thermal_inertia,

        beam_cross_section_area=beam_cross_section_area,
        beam_rho=beam_rho,
        beam_c=beam_c,
        beam_temperature_goal=beam_temperature_goal,
        protection_protected_perimeter=protection_protected_perimeter,
        protection_thickness=protection_thickness,
        protection_k=protection_k,
        protection_rho=protection_rho,
        protection_c=protection_c,

        iso834_time=iso834_time,
        iso834_temperature=iso834_temperature,
    )

    dict_pref_kwargs = dict(
        simulations=simulations,
        # steel_temp_failure=beam_temperature_goal,
        n_proc=n_proc,
        reliability_target=reliability_target,
        # select_fires_teq=select_fires_teq,
        # select_fires_teq_tol=select_fires_teq_tol,
        # is_deterministic=is_deterministic
    )

    list_mc_input_kwargs = []

    if simulations <= 2:
        x_ = dict_input_kwargs_static.copy()
        x_.update({"window_open_fraction": opening_fraction_mean,
                   "fire_load_density": qfd_mean,
                   "fire_spread_speed": (spread_ubound+spread_lbound)/2,
                   "beam_position": room_depth*0.75,
                   "nft_ubound": nft_mean,
                   "index": 0}, )
        list_mc_input_kwargs.append(x_)
    else:
        for i in range(0, simulations):
            x_ = dict_input_kwargs_static.copy()
            x_.update({"window_open_fraction": glaz_lhs[i],
                       "fire_load_density": qfd_lhs[i],
                       "fire_spread_speed": spread_lhs[i],
                       "beam_position": beam_lhs[i],
                       "nft_ubound": nft_lhs[i],
                       "index": i}, )
            list_mc_input_kwargs.append(x_)

    # list_mc_input_kwargs["window_open_fraction"] = glaz_lhs
    # dict_vars_0_["fire_load_density"] = qfd_lhs
    # dict_vars_0_["fire_spread_speed"] = spread_lhs
    # dict_vars_0_["beam_position"] = beam_lhs
    # dict_vars_0_["nft_ubound"] = nft_lhs
    # dict_vars_0_["index"] = np.arange(0, simulations, 1, int)

    df_input = df(list_mc_input_kwargs)
    df_input.set_index("index", inplace=True)

    return df_input, dict_pref_kwargs


def mc_inputs_generator_core2(
        simulations,
        # reliability_target,

        # time_start,
        # time_step,
        # time_end,

        # room_breadth,
        room_depth,
        # room_height,
        # room_window_width,
        # room_window_height,
        # room_wall_thermal_inertia,
        room_opening_fraction_lbound,
        room_opening_fraction_ubound,
        room_opening_fraction_std,
        room_opening_fraction_mean,
        room_opening_permanent_fraction,

        # fire_tlim,
        # fire_hrr_density,
        fire_qfd_std,
        fire_qfd_mean,
        fire_qfd_ubound,
        fire_qfd_lbound,
        fire_com_eff_lbound,
        fire_com_eff_ubound,
        fire_spread_lbound,
        fire_spread_ubound,
        fire_nft_mean,

        # beam_cross_section_area,
        # beam_rho,
        # beam_c,
        # beam_temperature_goal,
        # beam_protection_protected_perimeter,
        # beam_protection_thickness,
        # beam_protection_k,
        # beam_protection_rho,
        # beam_protection_c,
        beam_loc_ratio_lbound,
        beam_loc_ratio_ubound,

        # Derived
        # iso834_time,
        # iso834_temperature,

        # trivial
        index=False,
        **abundend_params
):

    # linear_distribution = lambda min, max, prob: ((max - min) * prob) + min
    def linear_distribution(min_, max_, rv_):
        return ((max_ - min_) * rv_) + min_

    # ==================================================================================================================
    # CHECKS
    # ==================================================================================================================

    # Fire duration has to be as long as travelling fire to travel through the entire floor
    # time_end = np.max([time_end, room_depth / fire_spread_lbound])

    # ==================================================================================================================
    # Distribution variables
    # ==================================================================================================================

    if simulations > 2:

        # lhs_mat = lhs(n=6, samples=simulations, criterion=dict_setting_vars["lhs_criterion"])
        lhs_mat = latin_hypercube_sampling(num_samples=simulations, num_arguments=6)

        # Near field standard deviation
        # fire_nft_mean = fire_nft_mean  # TFM near field temperature - Norm distribution - mean [C]
        std_nft = (1.939 - (np.log(fire_nft_mean) * 0.266)) * fire_nft_mean

        # Convert LHS probabilities to distribution invariants
        comb_lhs = linear_distribution(fire_com_eff_lbound, fire_com_eff_ubound, lhs_mat[:, 0])

        # Fuel load density
        # -----------------
        qfd_loc, qfd_scale = gumbel_parameter_converter(fire_qfd_mean, fire_qfd_std)
        qfd_lhs = gumbel_r_trunc_ppf(fire_qfd_lbound, fire_qfd_ubound, simulations, qfd_loc, qfd_scale) * comb_lhs
        np.random.shuffle(qfd_lhs)
        # DEPRECIATED 14 Aug 2018 - Following codes calculate gumbel parameters for qfd
        # qfd_scale = qfd_std * (6 ** 0.5) / np.pi
        # qfd_loc = qfd_mean - (0.57722 * qfd_scale)
        # qfd_dist = gumbel_r(loc=qfd_loc, scale=qfd_scale)
        # qfd_p_l, qfd_p_u = qfd_dist.cdf(qfd_lbound), qfd_dist.cdf(qfd_ubound)
        # qfd_lhs = gumbel_r(loc=qfd_loc, scale=qfd_scale).ppf(lhs_mat[:, 1]) * comb_lhs

        # Opening fraction factor (glazing fall-out fraction)
        # ---------------------------------------------------
        opening_fraction_mean_conv, opening_fraction_std_conv = lognorm_parameters_true_to_inv(room_opening_fraction_mean,
                                                                                               room_opening_fraction_std)
        glaz_lhs = 1 - lognorm_trunc_ppf(room_opening_fraction_lbound, room_opening_fraction_ubound, simulations,
                                         opening_fraction_std_conv, 0, np.exp(opening_fraction_mean_conv))
        np.random.shuffle(glaz_lhs)

        # count for permanent opening area
        glaz_lhs = [i * (1 - room_opening_permanent_fraction) + room_opening_permanent_fraction for i in glaz_lhs]

        # Beam location
        # -------------
        beam_lhs = linear_distribution(beam_loc_ratio_lbound, beam_loc_ratio_ubound, lhs_mat[:, 3]) * room_depth

        # Fire spread speed (travelling fire)
        # -----------------------------------
        spread_lhs = linear_distribution(fire_spread_lbound, fire_spread_ubound, lhs_mat[:, 4])

        # Near field temperature (travelling fire)
        # ----------------------------------------
        nft_lhs = norm(loc=fire_nft_mean, scale=std_nft).ppf(lhs_mat[:, 5])
        nft_lhs[nft_lhs > 1200] = 1200  # todo: reference?

    # ------------------------------------------------------------------------------------------------------------------
    # Create input kwargs for mc calculation
    # ------------------------------------------------------------------------------------------------------------------

    # if index:
    #     dict_input_kwargs_static['index'] = np.arange(0, simulations)

    list_mc_input_kwargs = []

    if simulations <= 2:
        x_ = {"window_open_fraction": room_opening_fraction_mean,
                   "fire_load_density": fire_qfd_mean,
                   "fire_spread_speed": (fire_spread_ubound + fire_spread_lbound) / 2,
                   "beam_position": room_depth*0.75,
                   "nft_ubound": fire_nft_mean,
                   "index": 0}
        list_mc_input_kwargs.append(x_)
    else:
        for i in range(0, int(simulations)):
            x_ = {"window_open_fraction": glaz_lhs[i],
                       "fire_load_density": qfd_lhs[i],
                       "fire_spread_speed": spread_lhs[i],
                       "beam_position": beam_lhs[i],
                       "nft_ubound": nft_lhs[i],
                       "index": i}
            list_mc_input_kwargs.append(x_)

    df_input = df(list_mc_input_kwargs)
    df_input.set_index("index", inplace=True)

    return df_input



# -*- coding: utf-8 -*-
import numpy as np
from scipy.interpolate import interp1d


def unprotected_steel_eurocode(
        time,
        temperature_ambient,
        perimeter_section,
        area_section,
        perimeter_box,
        density_steel,
        c_steel_T,
        h_conv,
        emissivity_resultant
):
    """
    SI UNITS FOR INPUTS AND OUTPUTS.
    :param time:
    :param temperature_ambient:
    :param perimeter_section:
    :param area_section:
    :param perimeter_box:
    :param density_steel:
    :param c_steel_T:
    :param h_conv:
    :param emissivity_resultant:
    :return:
    """

    # Create steel temperature change array s
    temperature_rate_steel = time * 0.
    temperature_steel = time * 0.
    heat_flux_net = time * 0.
    c_s = time * 0.

    k_sh = 0.9 * (perimeter_box / area_section) / (perimeter_section / area_section)  # BS EN 1993-1-2:2005 (e4.26a)
    F = perimeter_section
    V = area_section
    rho_s = density_steel
    h_c = h_conv
    sigma = 56.7e-9  # todo: what is this?
    epsilon = emissivity_resultant

    time_, temperature_steel[0], c_s[0] = iter(time), temperature_ambient[0], 0.
    next(time_)
    for i, v in enumerate(time_):
        i += 1
        T_f, T_s_ = temperature_ambient[i], temperature_steel[i-1]  # todo: steel specific heat
        c_s[i] = c_steel_T(temperature_steel[i - 1] + 273.15)

        # BS EN 1993-1-2:2005 (e4.25)
        a = h_c * (T_f - T_s_)
        b = sigma * epsilon * (np.power(T_f,4) - np.power(T_s_,4))
        c = k_sh * F / V / rho_s / c_s[i]
        d = time[i] - time[i-1]

        heat_flux_net[i] = a + b

        temperature_rate_steel[i] = c * (a + b) * d
        temperature_steel[i] = temperature_steel[i-1] + temperature_rate_steel[i]

    return temperature_steel, temperature_rate_steel, heat_flux_net, c_s


def protected_steel_eurocode(
        time,
        temperature_ambient,
        rho_steel,
        c_steel_T,
        area_steel_section,
        k_protection,
        rho_protection,
        c_protection,
        thickness_protection,
        perimeter_protected,
        is_terminate_peak=False
):
    """
    SI UNITS!
    This function calculate the temperature curve of protected steel section based on BS EN 1993-1-2:2005, Section 4
    . Ambient (fire) time-temperature data must be given, as well as the parameters specified below.
    :param time:                    {ndarray} [s]
    :param temperature_ambient:     {ndarray} [K]
    :param rho_steel:               {float} [kg/m3]
    :param c_steel_T:               {Func} [J/kg/K]
    :param area_steel_section:      {float} [m2]
    :param k_protection:            {float} [K/kg/m]
    :param rho_protection:          {float} [kg/m3]
    :param c_protection:            {float} [J/K/kg]
    :param thickness_protection:    {float} [m]
    :param perimeter_protected:     {float} [m]
    :param is_terminate_peak:       {bool} [-]              True will terminate and return values when first peak steel temperature is observed.
    :return time:                   {ndarray, float} [s]
    :return temperature_steel:      {ndarray, float} [K]
    :return data_all:               {Dict} [-]
    """
    # todo: 4.2.5.2 (2) - thermal properties for the insulation material
    # todo: revise BS EN 1993-1-2:2005, Clauses 4.2.5.2

    V = area_steel_section
    rho_a = rho_steel
    lambda_p = k_protection
    rho_p = rho_protection
    d_p = thickness_protection
    A_p = perimeter_protected
    c_p = c_protection

    temperature_steel = time * 0.
    temperature_rate_steel = time * 0.
    specific_heat_steel = time * 0.

    # Check time step <= 30 seconds. [BS EN 1993-1-2:2005, Clauses 4.2.5.2 (3)]
    # time_change = gradient(time)
    # if np.max(time_change) > 30.:
    # raise ValueError("Time step needs to be less than 30s: {0}".format(np.max(time)))

    temperature_steel[0] = temperature_ambient[0]  # initially, steel temperature is equal to ambient
    temperature_ambient_ = iter(temperature_ambient)  # skip the first item
    next(temperature_ambient_)  # skip the first item
    for i, T_g in enumerate(temperature_ambient_):
        i += 1  # actual index since the first item had been skipped.
        specific_heat_steel[i] = c_steel_T(temperature_steel[i - 1])

        # Steel temperature equations are from [BS EN 1993-1-2:2005, Clauses 4.2.5.2, Eq. 4.27]
        phi = (c_p * rho_p / specific_heat_steel[i] / rho_a) * d_p * A_p / V

        a = (lambda_p*A_p/V) / (d_p * specific_heat_steel[i] * rho_a)
        b = (T_g-temperature_steel[i-1]) / (1.+phi/3.)
        c = (np.exp(phi/10.)-1.) * (T_g-temperature_ambient[i-1])
        d = time[i] - time[i-1]

        temperature_rate_steel[i] = (a * b * d - c) / d  # deviated from e4.27, converted to rate [s-1]

        temperature_steel[i] = temperature_steel[i-1] + temperature_rate_steel[i] * d

        # NOTE: Steel temperature can be decrease at the begining, even the ambient temperature (fire) is hot. This is
        #       due to the factor 'phi' which intends to address the energy locked within the protection layer.
        #       The steel temperature is forced to be increased or remain as previous when ambient temperature and
        #       its previous temperature are all higher than the current calculated temperature.
        #       A better implementation is perhaps to use a 1-D heat transfer model.

        if temperature_steel[i] < min(temperature_steel[i-1], temperature_ambient[i]):
            temperature_rate_steel[i] = 0
            temperature_steel[i] = temperature_steel[i-1]

        if is_terminate_peak and temperature_rate_steel[i] < 0:
            data_all = {"temperature fire [K]": temperature_ambient,
                        "temperature rate steel [K/s]": temperature_rate_steel,
                        "specific heat steel [J/kg/K]": specific_heat_steel}
            return time, temperature_steel, data_all

    data_all = {
        "temperature steel [K]": temperature_steel,
        "temperature rate steel [K/s]": temperature_rate_steel,
        "specific heat steel [J/kg/K]": specific_heat_steel
    }

    return time, temperature_steel, data_all


def protected_steel_bs13381(
        time, temperature_ambient,
        lambda_pt_T, d_p, A_p, V, c_a_T, rho_a_T,
        time_ubound=10800, time_step_lbound=30.
):
    """
    :param time:                {ndarray}   [s]         Time array incorporating temperature_ambient, forming a time-temperature curve
    :param temperature_ambient: {ndarray}   [s]         Temperature array incorporating time, forming a time-temperature curve
    :param lambda_pt_T:         {func}      [W/m/K]     Thermal conductivity of the protective material (temperature dependent)
    :param d_p:                 {float}     [m]         Dry film thickness of reactive product
    :param A_p:                 {float}     [m]         Perimeter of protected
    :param V:                   {float}     [m2]        Area of steel cross section
    :param c_a_T:               {func}      [J/kg/K]    Specific heat capacity of steel (temperature dependent)
    :param rho_a_T:             {func}      [kg/m3]     Density of steel (temperature dependent)
    :param time_ubound:         {float}     [s]         The calculation time
    :param time_step_lbound:    {float}     [s]         The user-defined minimum time step, however, a smaller time step will be used with E.2
    :return:
    """

    # DEFINE FUNCTIONS: There are several functions defined in BS EN 13381-8:2013, ANNEX E, E.3. They are defined in
    # advance for simplification.
    # [BS EN 13381-8:2013, ANNEX E, Equation E.1]
    def _theta_at(lambda_pt, d_p, A_p, V, c_a, rho_a, theta_t, theta_at, dt):
        return (lambda_pt/d_p) * (A_p/V) * (1/c_a/rho_a) * (theta_t-theta_at) * dt

    # [BS EN 13381-8:2013, ANNEX E, Equation E.2]
    def _dt(c_a, rho_a, lambda_pt, d_p, A_p, V):
        return 0.8 * (c_a * rho_a / lambda_pt * d_p) * (V / A_p)

    # [BS EN 13381-8:2013, ANNEX E, Equation E.3]
    def _lambda_pt(d_p, V, A_p, c_a, rho_a, theta_t, theta_at, dt, d_theta_at):
        return d_p * (V / A_p) * c_a * rho_a * (1 / ((theta_t-theta_at)*dt)) * d_theta_at

    # [BS EN 13381-8:2013, ANNEX E, Equation E.4]
    def _theta_pt(theta_t, theta_t_, theta_at, theta_at_):
        return ((theta_t_+theta_t)/2. + (theta_at_+theta_at)/2.) / 2.

    # [BS EN 13381-8:2013, ANNEX E, Equation E.5]
    def _d_theta_at(c_a, rho_a, lambda_ave, d_p, A_p, V, theta_t, theta_at, dt):
        return (1/(c_a+rho_a)) * (lambda_ave/d_p) * (A_p / V) * (theta_t - theta_at) * dt

    # [BS EN 13381-8:2013, ANNEX E, Equation E.6]
    def _lambda_char(lambda_ave, K, sigma):
        return lambda_ave + K * sigma

    # Note: only equation E.1 and E.2 are currently being used as it is assumed the conductivity of protection layer
    #       is known.

    _temperature_gas = interp1d(time, temperature_ambient)

    # Instantiate output containers
    time_ = [0]
    time_rate = [0]
    temperature_steel = [temperature_ambient[0]]
    temperature_rate_steel = [0]
    conductivity_protection = [np.nan]

    # Start iterative calculation
    t = 0
    i = 0
    while t < time_ubound:
        i += 1

        # Determine conditions
        T_g = _temperature_gas(t)

        # Determine thermal properties
        c_a = c_a_T(T_g)
        rho_a = rho_a_T(T_g)
        lambda_pt = lambda_pt_T(T_g)
        conductivity_protection[i] = lambda_pt

        # Determine time step
        dt = _dt(c_a, rho_a, lambda_pt, d_p, A_p, V)
        dt = min([dt, time_step_lbound])
        time_rate.append(dt)

        # Determine current time
        time_.append(time_[i-1] + dt)
        t = time_[i]

        # Determine temperature change rate of the steel
        lambda_ave, theta_t, theta_at = lambda_pt, T_g, temperature_steel[i-1]
        temperature_rate_steel.append(_d_theta_at(c_a, rho_a, lambda_ave, d_p, A_p, V, theta_t, theta_at, dt))
        temperature_steel.append(temperature_steel[i-1] + temperature_rate_steel[i])

    # Convert lists to ndarray
    time_, time_rate = np.asarray(time_), np.asarray(time_rate)
    temperature_steel, temperature_rate_steel = np.asarray(temperature_steel), np.asarray(temperature_rate_steel)
    data_trivial = {""}

    return time_, temperature_steel, time_rate, temperature_rate_steel



# -*- coding: utf-8 -*-
import numpy as np
import itertools


def distribute_numbers_cartesian_product(n):
    b = []
    for v in n:
        a = np.linspace(0, 1, v+2)
        a = a[0:-2] + a[1]
        b.append(a)

    c = itertools.product(*b)
    c = list(c)
    c = np.asarray(c)

    return c


def permuatation_list(a, b):

    if isinstance(a[0], list):
        a = [[i] for i in a]

    if isinstance(b[0], list):
        b = [[i] for i in b]

    r = []
    for i in a:
        for j in b:
            r.append(i+j)


if __name__ == "__main__":
    # print(distribute_numbers_cartesian_product([9]*3))

    a = [[1],[2],[3]]

    b = [[11],[22],[33]]

    r = permuatation_list(a, b)

    print(r)



import xml.etree.ElementTree as ET
from collections import defaultdict

__all__ = ['XML2Dict']


class XML2Dict(object):
    def __init__(self, coding='UTF-8'):
        self._coding = coding

    def _parse_node(self, t):
        d = {t.tag: {} if t.attrib else None}  # the variable 'd' is the constructed target dictionary
        # 't.tag' if have values, it is the first layer of the dictionary
        children = list(t)  # The following recursive traverse processing tree, until the leaf node
        if children:  # Determine whether the node is empty, recursive boundary conditions
            dd = defaultdict(list)
            for dc in map(self._parse_node, children):  # recursive traverse processing tree
                for k, v in dc.iteritems():
                    dd[k].append(v)
            d = {t.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.iteritems()}}  # handle child node
        if t.attrib:  # handle attributes,prefix all of the stored attributes @
            d[t.tag].update(('@' + k, v) for k, v in t.attrib.iteritems())
        if t.text:
            text = t.text.strip().encode(self._coding)  # strip blank space
            if children or t.attrib:
                d[t.tag]['#text'] = text
            else:
                d[t.tag] = text  # the text value as t.tag
        return d

    def parse(self, xml_file):
        with open(xml_file, 'r') as fp:
            return self.fromstring(fp.read())

    def fromstring(self, xml_str):
        t = ET.fromstring(xml_str)
        return self._parse_node(t)


# -*- coding: utf-8 -*-
import matplotlib

matplotlib.use('agg')

import multiprocessing as mp
import os
import time
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns
import json
import copy
from tkinter import filedialog, Tk, StringVar
from scipy.interpolate import interp1d

from sfeprapy.dat.steel_carbon import Thermal
from sfeprapy.func.temperature_fires import parametric_eurocode1 as _fire_param
from sfeprapy.func.temperature_fires import standard_fire_iso834 as _fire_standard
from sfeprapy.func.tfm_alt import travelling_fire as _fire_travelling
from sfeprapy.time_equivalence_mc import calc_time_equiv_worker, mc_inputs_generator_core2, calc_time_equivalence

sns.set_style("ticks", {'axes.grid': True, })


def plot_dist(id_, df_input, path_input_file, headers):
    # headers = ['window_open_fraction', 'fire_load_density', 'fire_spread_speed', 'beam_position', 'temperature_max_near_field']
    # headers = ['WINDOW OPEN FRACTION [%]', 'FIRE LOAD DENSITY [MJ/m2]', 'FIRE SPREAD SPEED [m/s]', 'BEAM POSITION [m]', 'MAX. NEAR FIELD TEMPERATURE [C]']

    names = {'WINDOW OPEN FRACTION []': 'Ao',
             'FIRE LOAD DENSITY [MJ/m2]': 'qfd',
             'FIRE SPREAD SPEED [m/s]': 'spread',
             'BEAM POSITION [m]': 'beam_loc',
             'MAX. NEAR FIELD TEMPERATURE [C]': 'nft'}

    fig, ax = plt.subplots(figsize=(3.94, 2.76))  # (3.94, 2.76) for large and (2.5, 2) for small figure size

    for k, v in names.items():
        x = np.array(df_input[k].values, float)

        if k == 'MAX. NEAR FIELD TEMPERATURE [C]':
            x = x[x < 1200]

        sns.distplot(x, kde=False, rug=True, bins=50, ax=ax, norm_hist=True)

        # Normal plot parameters
        ax.set_ylabel('PDF')
        ax.set_xlabel(k)

        # Small simple plot parameters
        # ax.set_ylabel('')
        # ax.set_yticklabels([])
        # ax.set_xlabel('')

        plt.tight_layout()
        plt.savefig(
            os.path.join(os.path.dirname(path_input_file), '{} - dist - {}.png'.format(id_, v)),
            transparent=True,
            ppi=300
        )
        plt.cla()

    plt.clf()


def select_fires_teq(df_output):
    dict_fires = dict()

    # for i,v in enumerate(list_index_selected_fires) :
    args = None
    for i, v in df_output.iterrows():
        # get input arguments
        # args = df_input.loc[i].to_dict()
        args = v.to_dict()

        # get fire type
        # fire_type = int(df_output.loc[i]["FIRE TYPE [0:P, 1:T]"])
        fire_type = int(v["FIRE TYPE [0:P, 1:T]"])

        if fire_type == 0:  # parametric fire
            w, l, h = args["ROOM BREADTH [m]"], args["ROOM DEPTH [m]"], args["ROOM HEIGHT [m]"]
            inputs_parametric_fire = {
                "A_t": 2 * (w * l + w * h + h * l),
                "A_f": w * l,
                "A_v": args["WINDOW HEIGHT [m]"] * args["WINDOW WIDTH [m]"] * args["WINDOW OPEN FRACTION []"],
                "h_eq": args["WINDOW HEIGHT [m]"],
                "q_fd": args["FIRE LOAD DENSITY [MJ/m2]"] * 1e6,
                "lambda_": args["ROOM WALL THERMAL INERTIA [J/m2s1/2K]"] ** 2,
                # thermal inertia is used instead of k rho c.
                "rho": 1,  # see comment for lambda_
                "c": 1,  # see comment for lambda_
                "t_lim": args["TIME LIMITING []"],
                "time_end": args["FIRE DURATION [s]"],
                "time_step": args["TIME STEP [s]"],
                "time_start": args["TIME START [s]"],
                # "time_padding": (0, 0),
                "temperature_initial": 20 + 273.15,
            }
            tsec, temps = _fire_param(**inputs_parametric_fire)
        elif fire_type == 1:  # travelling fire
            inputs_travelling_fire = {
                "fire_load_density_MJm2": args["FIRE LOAD DENSITY [MJ/m2]"],
                "heat_release_rate_density_MWm2": args["FIRE HRR DENSITY [MW/m2]"],
                "length_compartment_m": args["ROOM DEPTH [m]"],
                "width_compartment_m": args["ROOM BREADTH [m]"],
                "fire_spread_rate_ms": args["FIRE SPREAD SPEED [m/s]"],
                "height_fuel_to_element_m": args["ROOM HEIGHT [m]"],
                "length_element_to_fire_origin_m": args["BEAM POSITION [m]"],
                "time_start_s": args["TIME START [s]"],
                "time_end_s": args["FIRE DURATION [s]"],
                "time_interval_s": args["TIME STEP [s]"],
                "nft_max_C": args["MAX. NEAR FIELD TEMPERATURE [C]"],
                "win_width_m": args["WINDOW WIDTH [m]"],
                "win_height_m": args["WINDOW HEIGHT [m]"],
                "open_fract": args["WINDOW OPEN FRACTION []"]
            }
            tsec, temps, hrr, r = _fire_travelling(**inputs_travelling_fire)
            temps += 273.15
        else:
            temps = 0
            print("FIRE TYPE UNKOWN.")

        dict_fires['FIRE {}'.format(str(i))] = temps - 273.15
        # plt.plot2(tsec/60., temps-273.15, alpha=.6)

    dict_fires["TIME [min]"] = np.arange(args["TIME START [s]"], args["FIRE DURATION [s]"], args["TIME STEP [s]"]) / 60.

    df_fires = pd.DataFrame(dict_fires)
    # list_names = ["TIME [min]"] + list_fire_name
    # df_fires = df_fires[list_names]

    # Save graphical plot to a .png file
    # ----------------------------------

    # Save numerical data to a .csv file
    # ----------------------------------
    # file_name = "{} - {}".format(os.path.basename(path_input_file).split('.')[0], __fn_fires_numerical)
    # df_fires.to_csv(os.path.join(os.path.dirname(path_input_file), file_name))
    return df_fires


def plot_figures(list_path_out_csv, plot_teq_xlim, reliability_target, path_work, figure_name='t_eq.png', **surplus_args):
    line_vertical, line_horizontal = 0, 0
    teq_fig, teq_ax = plt.subplots(figsize=(3.94, 2.76))
    teq_ax.set_xlim([0, plot_teq_xlim])

    for i, path_out_csv in enumerate(list_path_out_csv):
        df_result = pd.read_csv(path_out_csv)

        # Obtain variable of individual simulation case
        # ---------------------------------------------

        # path_input_file = list_path_input_file[i]
        # pref_ = list_pref[i]

        # Obtain some useful strings / directories / variables
        # ----------------------------------------------------

        id_ = os.path.basename(path_out_csv).replace('_out.csv', '').replace('_allout.csv', '')

        # Check 'deterministic'
        # ----------------------
        if len(df_result) <= 2:
            break

        # obtain x and y values for plot

        x = np.sort(df_result["TIME EQUIVALENCE [s]"].values) / 60.  # to minutes
        y = np.arange(1, len(x) + 1) / len(x)

        # plot the x, y

        plt.plot(x, y, label=id_)
        teq_ax.set_ylabel('Fractile')
        teq_ax.set_xlabel('Time [min]')
        teq_ax.set_xticks(ticks=np.arange(0, plot_teq_xlim + 0.001, 30))
        # plt.plot2(x, y, id_)
        # plt.format(**plt_format_)

        # plot horizontal and vertical lines (i.e. fractile and corresponding time euiqvalence period)

        if 0 < reliability_target < 1:
            line_horizontal = reliability_target
        elif reliability_target > 1:
            line_horizontal = 1 - 64.8 / reliability_target ** 2
        else:
            line_horizontal = 0

        if line_horizontal > 0:
            f_interp = interp1d(y, x)
            line_vertical = np.max((float(f_interp(line_horizontal)), line_vertical))

    if line_horizontal > 0:
        teq_ax.axhline(y=line_horizontal, c='grey')
        teq_ax.axvline(x=line_vertical, c='grey')
        teq_ax.text(
            x=line_vertical,
            y=teq_ax.get_ylim()[1],
            s="{:.0f}".format(line_vertical),
            va="bottom",
            ha="center",
            fontsize=9)

    teq_ax.legend().set_visible(True)
    teq_ax.legend(prop={'size': 7})
    plt.tight_layout()
    plt.savefig(
        os.path.join(path_work, figure_name),
        transparent=True,
        bbox_inches='tight',
        dpi=300
    )

    plt.clf()
    plt.close()


def csv_to_jsons():
    # get a list of dict()s representing different scenarios
    root = Tk()
    root.withdraw()
    folder_path = StringVar()

    path_input_file_csv = filedialog.askopenfile(title='Select Input Files', filetypes=[('csv', ['.csv'])])
    folder_path.set(path_input_file_csv)
    root.update()

    try:
        path_input_file_csv = os.path.realpath(path_input_file_csv.name)
    except AttributeError:
        print("terminated because no files are selected.")
        return 0

    path_work = os.path.dirname(path_input_file_csv)

    # csv to dict
    df_input_params = pd.read_csv(path_input_file_csv).set_index('PARAMETERS')
    dict_dict_input_params = df_input_params.to_dict()

    # dict to json (list)
    list_path_input_jsons = []
    for key, val in dict_dict_input_params.items():
        path_input_file_json = os.path.join(path_work, 'temp', key + '.json')

        if not val['is_live']:
            continue

        list_path_input_jsons.append(path_input_file_json)
        try:
            os.mkdir(os.path.dirname(path_input_file_json))
        except FileExistsError:
            pass

        with open(path_input_file_json, 'w') as f:
            json.dump(val, f)

    return path_work, list_path_input_jsons


def run():
    # ==================================================================================================================
    # Parse inputs from files into list of dict()s
    # ==================================================================================================================

    # get a list of dict()s representing different scenarios
    # requires path_work, list_path_input_json
    # from tkinter import filedialog, Tk, StringVar
    # root = Tk()
    # root.withdraw()
    # folder_path = StringVar()
    #
    # list_path_input_json = filedialog.askopenfiles(title='Select Input Files', )
    # folder_path.set(list_path_input_json)
    # root.update()
    #
    # list_path_input_json = [os.path.realpath(i.name) for i in list_path_input_json]


    # load csv input file
    path_work, list_path_input_json = csv_to_jsons()

    # load configuration parameters
    try:
        with open(os.path.join(path_work, 'config.json'), 'r') as f:
            dict_config_params = json.load(f)
    except FileExistsError:
        dict_config_params_default = {
            'n_proc': 3,
            'plot_teq_xlim': 120,
            'reliability_target': 0.8,
        }
        dict_config_params = dict_config_params_default

    list_input_file_names = [os.path.basename(i).split('.')[0] for i in list_path_input_json]

    # ==================================================================================================================
    # Plot all figures if no simulations to run
    # ==================================================================================================================

    if len(list_path_input_json) == 0:
        list_path_out_csv_all = []
        for f in os.listdir(os.path.join(path_work, 'temp')):
            if f.endswith("_out.csv"):
                list_path_out_csv_all.append(os.path.join(path_work, 'temp', f))

        # Plot individual t_eq
        for path_out_csv in list_path_out_csv_all:
            plot_figures(
                list_path_out_csv=[path_out_csv],
                plot_teq_xlim=dict_config_params['plot_teq_xlim'],
                reliability_target=dict_config_params['reliability_target'],
                path_work=os.path.join(path_work, 'temp'),
                figure_name=os.path.basename(path_out_csv).replace('_out.csv', '.png')
            )

        # Plot individual t_eq into one figure
        plot_figures(
            list_path_out_csv=list_path_out_csv_all,
            plot_teq_xlim=dict_config_params['plot_teq_xlim'],
            reliability_target=dict_config_params['reliability_target'],
            path_work=path_work,
            figure_name='t_eq.png'
        )

        # Plot combined t_eq with all results merged into one curve
        if len(list_path_out_csv_all) >= 1:
            list_pd_out_csv = []
            for path_out_csv in list_path_out_csv_all:
                # print(path_out_csv)
                list_pd_out_csv.append(pd.read_csv(path_out_csv, index_col='INDEX'))
            pd_out_all = pd.concat(list_pd_out_csv, axis=0, ignore_index=True)
            pd_out_all.sort_values("TIME EQUIVALENCE [s]", inplace=True)
            path_out_csv_merged = os.path.join(path_work, 'temp', '{}_allout.csv'.format(os.path.basename(path_work)))
            pd_out_all.to_csv(path_out_csv_merged)

            plot_figures(
                list_path_out_csv=[path_out_csv_merged],
                plot_teq_xlim=dict_config_params['plot_teq_xlim'],
                reliability_target=dict_config_params['reliability_target'],
                path_work=path_work,
                figure_name='t_eq_merged.png'
            )

        return 0

    list_dict_input_files = []
    for i in list_path_input_json:
        with open(i, 'r') as f:
            list_dict_input_files.append(json.load(f))

    # ==================================================================================================================
    # Spawn Monte Carlo parameters from dict() to DataFrame()s
    # ==================================================================================================================

    list_df_mc_params = []

    for i, dict_mc_params in enumerate(list_dict_input_files):

        # obtain user defined MC parameters if provided
        if 'path_mc_params_csv' in dict_mc_params:  # Read MC parameters. If USER provided MC parameters in a *.csv file
            # load MC parameters
            try:
                df_mc_params = pd.read_csv(dict_mc_params['path_mc_params_csv'])
            except FileNotFoundError:
                print('ERROR! *.csv file not found: {}'.format(dict_mc_params['path_mc_params_csv']))
                return -1
        else:  # Make MC parameters. If USER did not provide MC parameters in a *.csv file

            # only sample the defined probability for total number of simulation
            if 'simulations_multiplier' in dict_mc_params:
                dict_mc_params['simulations'] = int(dict_mc_params['simulations'] * dict_mc_params['simulations_multiplier']) + 1

            file_name_ = '{}_in.csv'.format(list_input_file_names[i])
            path_input_file_ = os.path.join(path_work, file_name_)

            if os.path.exists(path_input_file_) and os.path.isfile(path_input_file_) and False:
                # todo: user defined *.csv MC input file
                df_mc_params = pd.read_csv(path_input_file_).set_index('index', drop=True)
                n_sim = len(df_mc_params.index)

            else:
                df_mc_params = mc_inputs_generator_core2(index=True, **dict_mc_params)

        list_df_mc_params.append(df_mc_params)

    # Finalise MC Parameters in DataFrame() and Save to *.csv

    # make *_in.csv input parameter files
    path_work_dump = os.path.join(path_work, 'temp')
    try:
        os.mkdir(path_work_dump)
    except FileExistsError:
        pass

    # finalise and save to *_in.csv
    beam_c = Thermal().c()
    iso834_time, iso834_temperature = _fire_standard(np.arange(0, 6 * 60 * 60, 1), 273.15 + 20)
    list_path_in_csv = []
    for i, df_mc_params in enumerate(list_df_mc_params):
        dict_mc_params_static = {
            'time_step': list_dict_input_files[i]['time_step'],
            'time_limiting': list_dict_input_files[i]['fire_tlim'],
            'window_height': list_dict_input_files[i]['room_window_height'],
            'window_width': list_dict_input_files[i]['room_window_width'],
            'room_breadth': list_dict_input_files[i]['room_breadth'],
            'room_depth': list_dict_input_files[i]['room_depth'],
            'room_height': list_dict_input_files[i]['room_height'],
            'room_wall_thermal_inertia': list_dict_input_files[i]['room_wall_thermal_inertia'],
            'fire_hrr_density': list_dict_input_files[i]['fire_hrr_density'],
            'fire_duration': list_dict_input_files[i]['time_duration'],
            'beam_rho': list_dict_input_files[i]['beam_rho'],
            'beam_c': beam_c,
            'beam_cross_section_area': list_dict_input_files[i]['beam_cross_section_area'],
            'beam_temperature_goal': list_dict_input_files[i]['beam_temperature_goal'],
            'beam_loc_z': list_dict_input_files[i]['beam_loc_z'],
            'protection_k': list_dict_input_files[i]['beam_protection_k'],
            'protection_rho': list_dict_input_files[i]['beam_protection_rho'],
            'protection_c': list_dict_input_files[i]['beam_protection_c'],
            'protection_thickness': list_dict_input_files[i]['beam_protection_thickness'],
            'protection_protected_perimeter': list_dict_input_files[i]['beam_protection_protected_perimeter'],
            'iso834_time': iso834_time,
            'iso834_temperature': iso834_temperature,
            'time_start': list_dict_input_files[i]['time_start'],
        }
        # dict_mc_params_static = {j:[dict_mc_params_static[j]]*len(df_mc_params.index) for j in dict_mc_params_static}
        for key, val in dict_mc_params_static.items():
            list_val = []
            for i_ in range(len(df_mc_params.index)):
                list_val.append(val)
            dict_mc_params_static[key] = list_val

        dict_mc_params = {j: df_mc_params[j].values for j in df_mc_params}
        dict_mc_params.update(dict_mc_params_static)
        dict_mc_params['index'] = df_mc_params.index.values

        df_mc_params = pd.DataFrame.from_dict(dict_mc_params).set_index('index')

        file_name_ = '{}_in.csv'.format(list_input_file_names[i])
        path_input_file_ = os.path.join(path_work_dump, file_name_)

        try:
            df_mc_params.to_csv(path_input_file_)
            list_path_in_csv.append(path_input_file_)
        except PermissionError:
            print('WARNING! File save failed: {}'.format(path_input_file_))
            pass

    # load *.csv
    list_df_mc_params = [pd.read_csv(i).set_index('index', drop=False) for i in list_path_in_csv]
    # replace str with objects, e.g. list()
    for i, v in enumerate(list_df_mc_params):
        dict_ = v.to_dict(orient='list')
        dict_['beam_c'] = [beam_c for i in range(len(v.index))]
        dict_['iso834_time'] = [iso834_time for i in range(len(v.index))]
        dict_['iso834_temperature'] = [iso834_temperature for i in range(len(v.index))]
        v = pd.DataFrame.from_dict(dict_).set_index('index')
        list_df_mc_params[i] = copy.deepcopy(v)

    # convert DataFrame() to list of dict()s
    list_list_dict_mc_params = []
    for i, dict_mc_params in enumerate(list_df_mc_params):
        list_dict_mc_params = []
        for j, w in dict_mc_params.iterrows():
            x = w.to_dict()
            x['index'] = j
            list_dict_mc_params.append(x)
        list_list_dict_mc_params.append(list_dict_mc_params)

    # ==================================================================================================================
    # Main calculation
    # ==================================================================================================================

    list_path_out_csv = []

    for i, list_dict_mc_params in enumerate(list_list_dict_mc_params):

        print("CASE:", list_input_file_names[i])
        print("NO. OF THREADS:", dict_config_params['n_proc'])
        print("NO. OF SIMULATION:", len(list_dict_mc_params))

        if dict_config_params['n_proc'] == 1:
            results = []

            for dict_mc_params in list_dict_mc_params:
                results.append(calc_time_equivalence(**dict_mc_params))

        else:
            time_simulation_start = time.perf_counter()
            m = mp.Manager()
            q = m.Queue()
            p = mp.Pool(int(dict_config_params['n_proc']), maxtasksperchild=1000)
            jobs = p.map_async(calc_time_equiv_worker, [(dict_, q) for dict_ in list_dict_mc_params])
            count_total_simulations = len(list_dict_mc_params)
            n_steps = 24  # length of the progress bar
            while True:
                if jobs.ready():
                    time_simulation_consumed = time.perf_counter() - time_simulation_start
                    print("{}{} {:.1f}s ".format('█' * round(n_steps), '-' * round(0), time_simulation_consumed))
                    break
                else:
                    p_ = q.qsize() / count_total_simulations * n_steps
                    print("{}{} {:03.1f}%".format('█' * int(round(p_)), '-' * int(n_steps - round(p_)),
                                                  p_ / n_steps * 100),
                          end='\r')
                    time.sleep(1)
            p.close()
            p.join()
            results = jobs.get()

        # save to *.csv
        results = np.array(results)
        list_path_out_csv.append(os.path.join(path_work, 'temp', '{}_out.csv'.format(list_input_file_names[i])))
        df_output = pd.DataFrame({'TIME STEP [s]': results[:, 0],
                                  'TIME START [s]': results[:, 1],
                                  'TIME LIMITING []': results[:, 2],
                                  'WINDOW HEIGHT [m]': results[:, 3],
                                  'WINDOW WIDTH [m]': results[:, 4],
                                  'WINDOW OPEN FRACTION []': results[:, 5],
                                  'ROOM BREADTH [m]': results[:, 6],
                                  'ROOM DEPTH [m]': results[:, 7],
                                  'ROOM HEIGHT [m]': results[:, 8],
                                  'ROOM WALL THERMAL INERTIA [J/m2s1/2K]': results[:, 9],
                                  'FIRE LOAD DENSITY [MJ/m2]': results[:, 10],
                                  'FIRE HRR DENSITY [MW/m2]': results[:, 11],
                                  'FIRE SPREAD SPEED [m/s]': results[:, 12],
                                  'FIRE DURATION [s]': results[:, 13],
                                  'BEAM POSITION [m]': results[:, 14],
                                  'BEAM RHO [kg/m3]': results[:, 15],
                                  'BEAM C [-]': results[:, 16],
                                  'BEAM CROSS-SECTION AREA [m2]': results[:, 17],
                                  'BEAM FAILURE TEMPERATURE [C]': results[:, 18],
                                  'PROTECTION K [W/m/K]': results[:, 19],
                                  'PROTECTION RHO [kg/m3]': results[:, 20],
                                  'PROTECTION C OBJECT []': results[:, 21],
                                  'PROTECTION THICKNESS [m]': results[:, 22],
                                  'PROTECTION PERIMETER [m]': results[:, 23],
                                  'ISO834 TIME ARRAY [s]': results[:, 24],
                                  'ISO834 TEMPERATURE ARRAY [K]': results[:, 25],
                                  'MAX. NEAR FIELD TEMPERATURE [C]': results[:, 26],
                                  'SEEK ITERATION LIMIT []': results[:, 27],
                                  'SEEK PROTECTION THICKNESS UPPER BOUND [m]': results[:, 28],
                                  'SEEK PROTECTION THICKNESS LOWER BOUND [m]': results[:, 29],
                                  'SEEK BEAM FAILURE TEMPERATURE TOLERANCE [K]': results[:, 30],
                                  'INDEX': results[:, 31],
                                  'TIME EQUIVALENCE [s]': results[:, 32],
                                  'SEEK STATUS [0:Fail, 1:Success]': results[:, 33],
                                  'FIRE TYPE [0:P, 1:T]': results[:, 34],
                                  'SOUGHT BEAM TEMPERATURE [K]': results[:, 35],
                                  'SOUGHT BEAM PROTECTION THICKNESS [m]': results[:, 36],
                                  'SOUGHT ITERATIONS []': results[:, 37],
                                  'BEAM TEMPERATURE TO FIXED PROTECTION THICKNESS [K]': results[:, 38],
                                  'FIRE TIME ARRAY [s]': results[:, 39],
                                  'FIRE TEMPERATURE ARRAY [K]': results[:, 40],
                                  'OPENING FACTOR [m0.5]': results[:, 41]
                                  })
        df_output.set_index("INDEX", inplace=True)  # assign 'INDEX' column as DataFrame index
        df_output.sort_values('TIME EQUIVALENCE [s]', inplace=True)  # sort base on time equivalence
        df_output.to_csv(list_path_out_csv[-1])

        # export fires
        df_fires = select_fires_teq(df_output)
        df_fires.to_csv(os.path.join(path_work, 'temp', '{}_fires.csv'.format(list_input_file_names[i])))

    # ==================================================================================================================
    # Postprocessing - plot figures
    # ==================================================================================================================

    # Requirements
    # list_path_out_csv: a list of numeral output file path in *.csv format

    list_path_out_csv_all = []
    for f in os.listdir(os.path.join(path_work, 'temp')):
        if f.endswith("_out.csv"):
            list_path_out_csv_all.append(os.path.join(path_work, 'temp', f))

    # Plot individual t_eq
    for path_out_csv in list_path_out_csv_all:
        plot_figures(
            list_path_out_csv=[path_out_csv],
            plot_teq_xlim=dict_config_params['plot_teq_xlim'],
            reliability_target=dict_config_params['reliability_target'],
            path_work=os.path.join(path_work, 'temp'),
            figure_name=os.path.basename(path_out_csv).replace('_out.csv', '.png')
        )

    # Plot individual t_eq into one figure
    plot_figures(
        list_path_out_csv=list_path_out_csv_all,
        plot_teq_xlim=dict_config_params['plot_teq_xlim'],
        reliability_target=dict_config_params['reliability_target'],
        path_work=path_work,
        figure_name='t_eq.png'
    )

    # Plot combined t_eq with all results merged into one curve
    if len(list_path_out_csv_all) >= 0:
        list_pd_out_csv = []
        for path_out_csv in list_path_out_csv_all:
            # print(path_out_csv)
            list_pd_out_csv.append(pd.read_csv(path_out_csv, index_col='INDEX'))
        pd_out_all = pd.concat(list_pd_out_csv, axis=0, ignore_index=True)
        pd_out_all.sort_values("TIME EQUIVALENCE [s]", inplace=True)
        path_out_csv_merged = os.path.join(path_work, 'temp', '{}_allout.csv'.format(os.path.basename(path_work)))
        pd_out_all.to_csv(path_out_csv_merged)

        plot_figures(
            list_path_out_csv=[path_out_csv_merged],
            plot_teq_xlim=dict_config_params['plot_teq_xlim'],
            reliability_target=dict_config_params['reliability_target'],
            path_work=path_work,
            figure_name='t_eq_merged.png'
        )


if __name__ == '__main__':
    run()



import sys

from PySide2.QtWidgets import QDialog, QApplication
from PySide2.QtWidgets import QHBoxLayout, QVBoxLayout
from PySide2.QtWidgets import QLineEdit, QLabel, QPushButton
import PySide2.QtWidgets as QtWidgets
from PySide2 import QtGui, QtCore
import pandas
import os


class TabForm(QtWidgets.QTabWidget):
    """"""

    def __init__(self, parent=None):
        """Constructor"""
        super(TabForm, self).__init__(parent)

    def button_click(self):
        for i in self.list_qt_edit:
            print(i.text())
        print('hello')


class Form(QDialog):
    """"""

    def __init__(self, df_data, parent=None):
        """Constructor"""
        super(Form, self).__init__(parent)

        self.setWindowTitle('SfePraPy')

        layout_main = QtWidgets.QVBoxLayout()
        layout_param = QtWidgets.QFormLayout()

        list_qt_label = []
        self.list_qt_edit = []
        dict_validator = {'int': QtGui.QIntValidator(), 'float': QtGui.QDoubleValidator(), 'str': None}
        dict_fmt = {'int': '{:.0f}', 'float': '{}', 'str': '{}'}
        for i, r in df_data.iterrows():

            list_qt_label.append(QtWidgets.QLabel(r['DISPLAY NAME']))

            self.list_qt_edit.append(QLineEdit())

            self.list_qt_edit[-1].setValidator(dict_validator[r['TYPE']])
            self.list_qt_edit[-1].setText(dict_fmt[r['TYPE']].format(r['VALUE']))

            layout_param.addRow(list_qt_label[-1], self.list_qt_edit[-1])

        layout_main.addLayout(layout_param, stretch=1)

        button = QPushButton('&Submit')
        button.clicked.connect(self.button_click)

        layout_main.addWidget(button)
        self.setLayout(layout_main)

    def button_click(self):
        for i in self.list_qt_edit:
            print(i.text())


def gui_portal(df_data=None):

    if df_data is None:
        df_data = pandas.read_csv('gui.csv', index_col='INDEX')

    app = QApplication([])
    form = Form(df_data)
    form.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    gui_portal()



from __future__ import division
import sys
from matplotlib import pyplot as plt
from itertools import product
import numpy as np
from operator import itemgetter

if sys.version_info >= (3, 0):
    xrange = range

def get_text_position(text, ax=None):
    ax = ax or plt.gca()
    x, y = text.get_position()
    return (ax.xaxis.convert_units(x),
            ax.yaxis.convert_units(y))

def get_bboxes(objs, r, expand=(1.0, 1.0), ax=None):
    if ax is None:
        ax = plt.gca()
    return [i.get_window_extent(r).expanded(*expand).transformed(ax.transData.inverted()) for i in objs]

def get_midpoint(bbox):
    cx = (bbox.x0+bbox.x1)/2
    cy = (bbox.y0+bbox.y1)/2
    return cx, cy

def get_points_inside_bbox(x, y, bbox):
    x1, y1, x2, y2 = bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax
    x_in = np.logical_and(x>x1, x<x2)
    y_in = np.logical_and(y>y1, y<y2)
    return np.where(x_in & y_in)[0]

def get_renderer(fig):
    try:
        return fig.canvas.get_renderer()
    except AttributeError:
        return fig.canvas.renderer

def overlap_bbox_and_point(bbox, xp, yp):
    cx, cy = get_midpoint(bbox)

    dir_x = np.sign(cx-xp)
    dir_y = np.sign(cy-yp)

    if dir_x == -1:
        dx = xp - bbox.xmax
    elif dir_x == 1:
        dx = xp - bbox.xmin
    else:
        dx = 0

    if dir_y == -1:
        dy = yp - bbox.ymax
    elif dir_y == 1:
        dy = yp - bbox.ymin
    else:
        dy = 0
    return dx, dy

def move_texts(texts, delta_x, delta_y, bboxes=None, renderer=None, ax=None):
    if ax is None:
        ax = plt.gca()
    if bboxes is None:
        if renderer is None:
            r = get_renderer(ax.get_figure())
        else:
            r = renderer
        bboxes = get_bboxes(texts, r, (1, 1), ax=ax)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    for i, (text, dx, dy) in enumerate(zip(texts, delta_x, delta_y)):
        bbox = bboxes[i]
        x1, y1, x2, y2 = bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax
        if x1 + dx < xmin:
            dx = 0
        if x2 + dx > xmax:
            dx = 0
        if y1 + dy < ymin:
            dy = 0
        if y2 + dy > ymax:
            dy = 0

        x, y = get_text_position(text, ax=ax)
        newx = x + dx
        newy = y + dy
        text.set_position((newx, newy))

def optimally_align_text(x, y, texts, expand=(1., 1.), add_bboxes=[],
                         renderer=None, ax=None,
                         direction='xy'):
    """
    For all text objects find alignment that causes the least overlap with
    points and other texts and apply it
    """
    if ax is None:
        ax = plt.gca()
    if renderer is None:
        r = get_renderer(ax.get_figure())
    else:
        r = renderer
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    bboxes = get_bboxes(texts, r, expand, ax=ax)
    if 'x' not in direction:
        ha = ['']
    else:
        ha = ['center', 'left', 'right']
    if 'y' not in direction:
        va = ['']
    else:
        va = ['bottom', 'top', 'center']
    alignment = list(product(ha, va))
#    coords = np.array(zip(x, y))
    for i, text in enumerate(texts):
#        tcoords = np.array(text.get_position()).T
#        nonself_coords = coords[~np.all(coords==tcoords, axis=1)]
#        nonself_x, nonself_y = np.split(nonself_coords, 2, axis=1)
        counts = []
        for h, v in alignment:
            if h:
                text.set_ha(h)
            if v:
                text.set_va(v)
            bbox = text.get_window_extent(r).expanded(*expand).\
                                       transformed(ax.transData.inverted())
            c = len(get_points_inside_bbox(x, y, bbox))
            intersections = [bbox.intersection(bbox, bbox2) for bbox2 in
                             bboxes+add_bboxes]
            intersections = sum([abs(b.width*b.height) if b is not None else 0
                                 for b in intersections])
            # Check for out-of-axes position
            bbox = text.get_window_extent(r).transformed(ax.transData.inverted())
            x1, y1, x2, y2 = bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax
            if x1 < xmin or x2 > xmax or y1 < ymin or y2 > ymax:
                axout = 1
            else:
                axout = 0
            counts.append((axout, c, intersections))
        a, value = min(enumerate(counts), key=itemgetter(1))
        if 'x' in direction:
            text.set_ha(alignment[a][0])
        if 'y' in direction:
            text.set_va(alignment[a][1])
        bboxes[i] = text.get_window_extent(r).expanded(*expand).\
                                       transformed(ax.transData.inverted())
    return texts

def repel_text(texts, renderer=None, ax=None, expand=(1.2, 1.2),
               only_use_max_min=False, move=False):
    """
    Repel texts from each other while expanding their bounding boxes by expand
    (x, y), e.g. (1.2, 1.2) would multiply width and height by 1.2.
    Requires a renderer to get the actual sizes of the text, and to that end
    either one needs to be directly provided, or the axes have to be specified,
    and the renderer is then got from the axes object.
    """
    if ax is None:
        ax = plt.gca()
    if renderer is None:
        r = get_renderer(ax.get_figure())
    else:
        r = renderer
    bboxes = get_bboxes(texts, r, expand, ax=ax)
    xmins = [bbox.xmin for bbox in bboxes]
    xmaxs = [bbox.xmax for bbox in bboxes]
    ymaxs = [bbox.ymax for bbox in bboxes]
    ymins = [bbox.ymin for bbox in bboxes]

    overlaps_x = np.zeros((len(bboxes), len(bboxes)))
    overlaps_y = np.zeros_like(overlaps_x)
    overlap_directions_x = np.zeros_like(overlaps_x)
    overlap_directions_y = np.zeros_like(overlaps_y)
    for i, bbox1 in enumerate(bboxes):
        overlaps = get_points_inside_bbox(xmins*2+xmaxs*2, (ymins+ymaxs)*2,
                                             bbox1) % len(bboxes)
        overlaps = np.unique(overlaps)
        for j in overlaps:
            bbox2 = bboxes[j]
            x, y = bbox1.intersection(bbox1, bbox2).size
            overlaps_x[i, j] = x
            overlaps_y[i, j] = y
            direction = np.sign(bbox1.extents - bbox2.extents)[:2]
            overlap_directions_x[i, j] = direction[0]
            overlap_directions_y[i, j] = direction[1]

    move_x = overlaps_x*overlap_directions_x
    move_y = overlaps_y*overlap_directions_y

    delta_x = move_x.sum(axis=1)
    delta_y = move_y.sum(axis=1)

    q = np.sum(np.abs(delta_x) + np.abs(delta_y))
    if move:
        move_texts(texts, delta_x, delta_y, bboxes, ax=ax)
    return delta_x, delta_y, q

def repel_text_from_bboxes(add_bboxes, texts, renderer=None, ax=None,
                           expand=(1.2, 1.2), only_use_max_min=False,
                           move=False):
    """
    Repel texts from other objects' bboxes while expanding their (texts')
    bounding boxes by expand (x, y), e.g. (1.2, 1.2) would multiply width and
    height by 1.2.
    Requires a renderer to get the actual sizes of the text, and to that end
    either one needs to be directly provided, or the axes have to be specified,
    and the renderer is then got from the axes object.
    """
    if ax is None:
        ax = plt.gca()
    if renderer is None:
        r = get_renderer(ax.get_figure())
    else:
        r = renderer

    bboxes = get_bboxes(texts, r, expand, ax=ax)

    overlaps_x = np.zeros((len(bboxes), len(add_bboxes)))
    overlaps_y = np.zeros_like(overlaps_x)
    overlap_directions_x = np.zeros_like(overlaps_x)
    overlap_directions_y = np.zeros_like(overlaps_y)

    for i, bbox1 in enumerate(bboxes):
        for j, bbox2 in enumerate(add_bboxes):
            try:
                x, y = bbox1.intersection(bbox1, bbox2).size
                direction = np.sign(bbox1.extents - bbox2.extents)[:2]
                overlaps_x[i, j] = x
                overlaps_y[i, j] = y
                overlap_directions_x[i, j] = direction[0]
                overlap_directions_y[i, j] = direction[1]
            except AttributeError:
                pass

    move_x = overlaps_x*overlap_directions_x
    move_y = overlaps_y*overlap_directions_y

    delta_x = move_x.sum(axis=1)
    delta_y = move_y.sum(axis=1)

    q = np.sum(np.abs(delta_x) + np.abs(delta_y))
    if move:
        move_texts(texts, delta_x, delta_y, bboxes, ax=ax)
    return delta_x, delta_y, q

def repel_text_from_points(x, y, texts, renderer=None, ax=None,
                           expand=(1.2, 1.2), move=False):
    """
    Repel texts from all points specified by x and y while expanding their
    (texts'!) bounding boxes by expandby  (x, y), e.g. (1.2, 1.2)
    would multiply both width and height by 1.2.
    Requires a renderer to get the actual sizes of the text, and to that end
    either one needs to be directly provided, or the axes have to be specified,
    and the renderer is then got from the axes object.
    """
    assert len(x) == len(y)
    if ax is None:
        ax = plt.gca()
    if renderer is None:
        r = get_renderer(ax.get_figure())
    else:
        r = renderer
    bboxes = get_bboxes(texts, r, expand, ax=ax)

    move_x = np.zeros((len(bboxes), len(x)))
    move_y = np.zeros((len(bboxes), len(x)))
    for i, bbox in enumerate(bboxes):
        xy_in = get_points_inside_bbox(x, y, bbox)
        for j in xy_in:
            xp, yp = x[j], y[j]
            dx, dy = overlap_bbox_and_point(bbox, xp, yp)

            move_x[i, j] = dx
            move_y[i, j] = dy

    delta_x = move_x.sum(axis=1)
    delta_y = move_y.sum(axis=1)
    q = np.sum(np.abs(delta_x) + np.abs(delta_y))
    if move:
        move_texts(texts, delta_x, delta_y, bboxes, ax=ax)
    return delta_x, delta_y, q

def repel_text_from_axes(texts, ax=None, bboxes=None, renderer=None,
                         expand=None):
    if ax is None:
        ax = plt.gca()
    if renderer is None:
        r = get_renderer(ax.get_figure())
    else:
        r = renderer
    if expand is None:
        expand = (1, 1)
    if bboxes is None:
        bboxes = get_bboxes(texts, r, expand=expand, ax=ax)
    xmin, xmax = ax.get_xlim()
    ymin, ymax = ax.get_ylim()
    for i, bbox in enumerate(bboxes):
        x1, y1, x2, y2 = bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax
        dx, dy = 0, 0
        if x1 < xmin:
            dx = xmin - x1
        if x2 > xmax:
            dx = xmax - x2
        if y1 < ymin:
            dy = ymin - y1
        if y2 > ymax:
            dy = ymax - y2
        if dx or dy:
            x, y = get_text_position(texts[i], ax=ax)
            newx, newy = x + dx, y + dy
            texts[i].set_position((newx, newy))
    return texts

def float_to_tuple(a):
    try:
        a = float(a)
        return (a, a)
    except TypeError:
        assert len(a)==2
        assert all([bool(i) for i in a])
        return a

def adjust_text(texts, x=None, y=None, add_objects=None, ax=None,
                expand_text=(1.2, 1.2), expand_points=(1.2, 1.2),
                expand_objects=(1.2, 1.2), expand_align=(0.9, 0.9),
                autoalign='xy',  va='center', ha='center',
                force_text=0.5, force_points=0.5, force_objects=0.5,
                lim=100, precision=0,
                only_move={}, text_from_text=True, text_from_points=True,
                save_steps=False, save_prefix='', save_format='png',
                add_step_numbers=True, draggable=True, on_basemap=False,
                *args, **kwargs):
    """
    Iteratively adjusts the locations of texts. First moves all texts that are
    outside the axes limits inside. Then in each iteration moves all texts away
    from each other and from points. In the end hides texts and substitutes
    them with annotations to link them to the respective points.

    Args:
        texts (list): a list of text.Text objects to adjust
        x (seq): x-coordinates of points to repel from; if not provided only
            uses text coordinates
        y (seq): y-coordinates of points to repel from; if not provided only
            uses text coordinates
        add_objects (list): a list of additional matplotlib objects to avoid;
            they must have a .get_window_extent() method
        ax (obj): axes object with the plot; if not provided is determined by
            plt.gca()
        expand_text (seq): a tuple/list/... with 2 numbers (x, y) to expand
            texts when repelling them from each other; default (1.2, 1.2)
        expand_points (seq): a tuple/list/... with 2 numbers (x, y) to expand
            texts when repelling them from points; default (1.2, 1.2)
        expand_objects (seq): a tuple/list/... with 2 numbers (x, y) to expand
            texts when repelling them from points; default (1.2, 1.2)
        expand_align (seq): a tuple/list/... with 2 numbers (x, y) to expand
            texts when autoaligning texts; default (1., 1.)
        autoalign: If 'xy', the best alignment of all texts will be
            determined in all directions automatically before running the
            iterative adjustment; if 'x' will only align horizontally, if 'y' -
            vertically; overrides va and ha
        va (str): vertical alignment of texts
        ha (str): horizontal alignment of texts
        force_text (float): the repel force from texts is multiplied by this
            value; default 0.5
        force_points (float): the repel force from points is multiplied by this
            value; default 0.5
        force_objects (float): same as other forces, but for repelling
            additional objects
        lim (int): limit of number of iterations
        precision (float): up to which sum of all overlaps along both x and y
            to iterate; may need to increase for complicated situations;
            default 0, so no overlaps with anything.
        only_move (dict): a dict to restrict movement of texts to only certain
            axis. Valid keys are 'points' and 'text', for each of them valid
            values are 'x', 'y' and 'xy'. This way you can forbid moving texts
            along either of the axes due to overlaps with points, but let it
            happen if there is an overlap with texts: only_move={'points':'y',
            'text':'xy'}. Default: None, so everything is allowed.
        text_from_text (bool): whether to repel texts from each other; default
            True
        text_from_points (bool): whether to repel texts from points; default
            True; can helpful to switch of in extremely crouded plots
        save_steps (bool): whether to save intermediate steps as images;
            default False
        save_prefix (str): a path and/or prefix to the saved steps; default ''
        save_format (str): a format to save the steps into; default 'png
        *args and **kwargs: any arguments will be fed into plt.annotate after
            all the optimization is done just for plotting
        add_step_numbers (bool): whether to add step numbers as titles to the
            images of saving steps
        draggable (bool): whether to make the annotations draggable; default
            True
        on_basemap (bool): whether your plot uses the basemap library, stops
            labels going over the edge of the map; default False
    """
    if ax is None:
        ax = plt.gca()
    r = get_renderer(ax.get_figure())
    orig_xy = [get_text_position(text, ax=ax) for text in texts]
    orig_x = [xy[0] for xy in orig_xy]
    orig_y = [xy[1] for xy in orig_xy]
    force_objects = float_to_tuple(force_objects)
    force_text = float_to_tuple(force_text)
    force_points = float_to_tuple(force_points)
    if x is None:
        if y is None:
            x, y = orig_x, orig_y
        else:
            raise ValueError('Please specify both x and y, or neither')
    if y is None:
        raise ValueError('Please specify both x and y, or neither')
    if add_objects is None:
        text_from_objects = False
        add_bboxes = []
    else:
        try:
            add_bboxes = get_bboxes(add_objects, r, ax=ax)
        except:
            raise ValueError("Can't get bounding boxes from add_objects - is'\
                             it a flat list of matplotlib objects?")
            return
        text_from_objects = True
    for text in texts:
        text.set_va(va)
        text.set_ha(ha)
    if save_steps:
        if add_step_numbers:
            plt.title('0a')
        plt.savefig(save_prefix+'0a.'+save_format, format=save_format)
    elif on_basemap:
        ax.draw(r)

    if autoalign:
        if autoalign is not True:
            texts = optimally_align_text(x, y, texts, expand=expand_align,
                                         add_bboxes=add_bboxes,
                                         direction=autoalign, renderer=r,
                                         ax=ax)
        else:
            texts = optimally_align_text(orig_x, orig_y, texts,
                                         expand=expand_align,
                                         direction='xy',
                                         add_bboxes=add_bboxes, renderer=r,
                                         ax=ax)

    if save_steps:
        if add_step_numbers:
            plt.title('0b')
        plt.savefig(save_prefix+'0b.'+save_format, format=save_format)
    elif on_basemap:
        ax.draw(r)

    texts = repel_text_from_axes(texts, ax, renderer=r, expand=expand_points)
    history = [np.inf]*5
    for i in xrange(lim):
        q1, q2 = np.inf, np.inf

        if text_from_text:
            d_x_text, d_y_text, q1 = repel_text(texts, renderer=r, ax=ax,
                                                expand=expand_text)
        else:
            d_x_text, d_y_text, q1 = [0]*len(texts), [0]*len(texts), 0

        if text_from_points:
            d_x_points, d_y_points, q2 = repel_text_from_points(x, y, texts,
                                                   ax=ax, renderer=r,
                                                   expand=expand_points)
        else:
            d_x_points, d_y_points, q2 = [0]*len(texts), [0]*len(texts), 0

        if text_from_objects:
            d_x_objects, d_y_objects, q3 = repel_text_from_bboxes(add_bboxes,
                                                                  texts,
                                                             ax=ax, renderer=r,
                                                         expand=expand_objects)
        else:
            d_x_objects, d_y_objects, q3 = [0]*len(texts), [0]*len(texts), 0

        if only_move:
            if 'text' in only_move:
                if 'x' not in only_move['text']:
                    d_x_text = np.zeros_like(d_x_text)
                if 'y' not in only_move['text']:
                    d_y_text = np.zeros_like(d_y_text)
            if 'points' in only_move:
                if 'x' not in only_move['points']:
                    d_x_points = np.zeros_like(d_x_points)
                if 'y' not in only_move['points']:
                    d_y_points = np.zeros_like(d_y_points)
            if 'objects' in only_move:
                if 'x' not in only_move['objects']:
                    d_x_objects = np.zeros_like(d_x_objects)
                if 'y' not in only_move['objects']:
                    d_y_objects = np.zeros_like(d_y_objects)
        dx = (np.array(d_x_text) * force_text[0] +
              np.array(d_x_points) * force_points[0] +
              np.array(d_x_objects) * force_objects[0])
        dy = (np.array(d_y_text) * force_text[1] +
              np.array(d_y_points) * force_points[1] +
              np.array(d_y_objects) * force_objects[1])
        q = round(q1+q2+q3, 5)
        if q > precision and q < np.max(history):
            history.pop(0)
            history.append(q)
            move_texts(texts, dx, dy,
                       bboxes = get_bboxes(texts, r, (1, 1)), ax=ax)
            if save_steps:
                if add_step_numbers:
                    plt.title(i+1)
                plt.savefig(save_prefix+str(i+1)+'.'+save_format,
                            format=save_format)
            elif on_basemap:
                ax.draw(r)
        else:
            break

    for j, text in enumerate(texts):
        a = ax.annotate(text.get_text(), xy = (orig_xy[j]),
                    xytext=text.get_position(), *args, **kwargs)
        a.__dict__.update(text.__dict__)
        if draggable:
            a.draggable()
        texts[j].remove()
    if save_steps:
        if add_step_numbers:
            plt.title(i+1)
        plt.savefig(save_prefix+str(i+1)+'.'+save_format, format=save_format)
    elif on_basemap:
        ax.draw(r)

    return i+1


from PySide2.QtWidgets import QTableView, QVBoxLayout, QApplication
from PySide2.QtCore import QAbstractItemModel

from helper import UsesQApplication
import unittest


class VirtualList(QAbstractItemModel):
    def __getitem__(self, index):
        self._getItemCalled = True
        pass

    def rowCount(self, parent):
        return 5000

    def columnCount(self, parent):
        return 3

    def index(self, row, column, parent):
        return self.createIndex(row, column)

    def parent(self, index):
        return QModelIndex()

    def data(self, index, role):
        return "(%i, %i)" % (index.row(), index.column())


class TestQAbstractItemModel(UsesQApplication):
    def testSetModel(self):
        model = VirtualList()
        model._getItemCalled = False
        table = QTableView()
        table.setModel(model)
        table.show()
        self.assertFalse(model._getItemCalled)


if __name__ == "__main__":
    unittest.main()